From 3200b96bd42cc279634575153ab8a94020ac37e2 Mon Sep 17 00:00:00 2001
From: Marina Polyakova <m.polyakova@postgrespro.ru>
Date: Tue, 7 Aug 2018 13:30:13 +0300
Subject: [PATCH v10 4/4] Pgbench errors and serialization/deadlock retries

Client's run is aborted only in case of a serious error, for example, the
connection with the backend was lost. Otherwise if the execution of SQL or meta
command fails, the client's run continues normally until the end of the current
script execution (it is assumed that one transaction script contains only one
transaction).

Transactions with serialization or deadlock failures are rolled back and
repeated until they complete successfully or reach the maximum number of tries
(specified by the --max-tries option) / the maximum time of tries (specified by
the --latency-limit option). These options can be combined together; but if
none of them are used, the default value for the option --max-tries is set to 1
and failed transactions are not retried at all. If the last transaction run
fails, this transaction will be reported as failed, and the client variables
will be set as they were before the first run of this transaction.

If there're retries and/or errors their statistics are printed in the progress,
in the transaction / aggregation logs and in the end with other results (all and
for each script). A transaction error is reported here only if the last try of
this transaction fails. Also retries and errors are printed per-command with
average latencies if you use the appropriate benchmarking option
(--report-per-command, -r). If you want to group errors by basic types
(serialization errors / deadlock errors / other SQL errors / errors in meta
commands), use the option --errors-detailed.

If you want to distinguish between failures or errors by type (including which
limit for retries was violated and how far it was exceeded for the
serialization/deadlock errors), use the pgbench debugging output created with
the option --debug-fails or --debug. The first option is recommended for this
purpose because with the second option the debugging output can be very large.
---
 doc/src/sgml/ref/pgbench.sgml                |  421 +++++++++-
 src/bin/pgbench/pgbench.c                    | 1160 +++++++++++++++++++++++---
 src/bin/pgbench/t/001_pgbench_with_server.pl |  372 ++++++++-
 src/bin/pgbench/t/002_pgbench_no_server.pl   |    5 +
 src/fe_utils/conditional.c                   |   16 +-
 src/include/fe_utils/conditional.h           |    2 +
 6 files changed, 1756 insertions(+), 220 deletions(-)

diff --git a/doc/src/sgml/ref/pgbench.sgml b/doc/src/sgml/ref/pgbench.sgml
index 88cf8b3..12d1120 100644
--- a/doc/src/sgml/ref/pgbench.sgml
+++ b/doc/src/sgml/ref/pgbench.sgml
@@ -55,16 +55,20 @@ number of clients: 10
 number of threads: 1
 number of transactions per client: 1000
 number of transactions actually processed: 10000/10000
+maximum number of tries: 1
 tps = 85.184871 (including connections establishing)
 tps = 85.296346 (excluding connections establishing)
 </screen>
 
   The first six lines report some of the most important parameter
-  settings.  The next line reports the number of transactions completed
+  settings.  The seventh line reports the number of transactions completed
   and intended (the latter being just the product of number of clients
   and number of transactions per client); these will be equal unless the run
-  failed before completion.  (In <option>-T</option> mode, only the actual
-  number of transactions is printed.)
+  failed before completion or some SQL/meta command(s) failed.  (In
+  <option>-T</option> mode, only the actual number of transactions is printed.)
+  The next line reports the maximum number of tries for transactions with
+  serialization and/or deadlock failures (see <xref linkend="errors-and-retries"
+  endterm="errors-and-retries-title"/> for more information)
   The last two lines report the number of transactions per second,
   figured with and without counting the time to start database sessions.
  </para>
@@ -453,6 +457,17 @@ pgbench <optional> <replaceable>options</replaceable> </optional> <replaceable>d
         at all. They are counted and reported separately as
         <firstterm>skipped</firstterm>.
        </para>
+       <para>
+        The transaction with serialization or deadlock failure can be retried if
+        the total time of all its tries is less than
+        <replaceable>limit</replaceable> ms. This option can be combined with
+        the option <option>--max-tries</option> which limits the total number of
+        transaction tries. If none of them is used, the default value for the
+        option <option>--max-tries</option> is set to 1 and failed transactions
+        are not retried at all. See <xref linkend="errors-and-retries"
+        endterm="errors-and-retries-title"/> for more information about retrying
+        failed transactions.
+       </para>
        </listitem>
      </varlistentry>
 
@@ -513,22 +528,34 @@ pgbench <optional> <replaceable>options</replaceable> </optional> <replaceable>d
         Show progress report every <replaceable>sec</replaceable> seconds.  The report
         includes the time since the beginning of the run, the TPS since the
         last report, and the transaction latency average and standard
-        deviation since the last report.  Under throttling (<option>-R</option>),
-        the latency is computed with respect to the transaction scheduled
-        start time, not the actual transaction beginning time, thus it also
-        includes the average schedule lag time.
+        deviation since the last report.  If any transactions have received an
+        error in the SQL or meta command since the last report, they are also
+        reported as failed.  Under throttling (<option>-R</option>), the latency
+        is computed with respect to the transaction scheduled start time, not
+        the actual transaction beginning time, thus it also includes the average
+        schedule lag time.  If any transactions have been rolled back and
+        retried after a serialization/deadlock failure since the last report,
+        the report includes the number of such transactions and the sum of all
+        retries. Use the options <option>--max-tries</option> and/or
+        <option>--latency-limit</option> to enable transactions retries after
+        serialization/deadlock failures.
        </para>
       </listitem>
      </varlistentry>
 
      <varlistentry>
       <term><option>-r</option></term>
-      <term><option>--report-latencies</option></term>
+      <term><option>--report-per-command</option></term>
       <listitem>
        <para>
-        Report the average per-statement latency (execution time from the
-        perspective of the client) of each command after the benchmark
-        finishes.  See below for details.
+        Report the following statistics for each command after the benchmark
+        finishes: the average per-statement latency (execution time from the
+        perspective of the client), the number of errors and the number of
+        retries after serialization or deadlock failures in this command.  The
+        report displays retry statistics only if the maximum number of tries for
+        transactions is more than 1 (<option>--max-tries</option>) and/or the
+        maximum time of tries for transactions is used
+        (<option>--latency-limit</option>).  See below for details.
        </para>
       </listitem>
      </varlistentry>
@@ -657,6 +684,49 @@ pgbench <optional> <replaceable>options</replaceable> </optional> <replaceable>d
      </varlistentry>
 
      <varlistentry>
+      <term><option>--errors-detailed</option></term>
+      <listitem>
+       <para>
+        Report errors in per-transaction and aggregation logs, as well as in the
+        main and per-script reports, grouped by the following types:
+        <itemizedlist>
+          <listitem>
+            <para>
+              serialization errors;
+            </para>
+          </listitem>
+          <listitem>
+            <para>
+              deadlock errors;
+            </para>
+          </listitem>
+          <listitem>
+            <para>
+              other SQL errors;
+            </para>
+          </listitem>
+          <listitem>
+            <para>
+              errors in meta commands.
+            </para>
+          </listitem>
+        </itemizedlist>
+       </para>
+      </listitem>
+     </varlistentry>
+
+     <varlistentry>
+      <term><option>--debug-fails</option></term>
+      <listitem>
+       <para>
+        Print debugging output only for errors, serialization/deadlock failures
+        and retries. See <xref linkend="errors-and-retries"
+        endterm="errors-and-retries-title"/> for more information.
+       </para>
+      </listitem>
+     </varlistentry>
+
+     <varlistentry>
       <term><option>--log-prefix=<replaceable>prefix</replaceable></option></term>
       <listitem>
        <para>
@@ -667,6 +737,22 @@ pgbench <optional> <replaceable>options</replaceable> </optional> <replaceable>d
      </varlistentry>
 
      <varlistentry>
+      <term><option>--max-tries=<replaceable>number_of_tries</replaceable></option></term>
+      <listitem>
+       <para>
+        Set the maximum number of tries for transactions with
+        serialization/deadlock failures. This option can be combined with the
+        option <option>--latency-limit</option> which limits the total time of
+        transaction tries. If none of them is used, the default value for the
+        option <option>--max-tries</option> is set to 1 and failed transactions
+        are not retried at all. See <xref linkend="errors-and-retries"
+        endterm="errors-and-retries-title"/> for more information about retrying
+        failed transactions.
+       </para>
+      </listitem>
+     </varlistentry>
+
+     <varlistentry>
       <term><option>--progress-timestamp</option></term>
       <listitem>
        <para>
@@ -807,8 +893,8 @@ pgbench <optional> <replaceable>options</replaceable> </optional> <replaceable>d
  <refsect1>
   <title>Notes</title>
 
- <refsect2>
-  <title>What is the <quote>Transaction</quote> Actually Performed in <application>pgbench</application>?</title>
+ <refsect2 id="transactions-and-scripts">
+  <title id="transactions-and-scripts-title">What is the <quote>Transaction</quote> Actually Performed in <application>pgbench</application>?</title>
 
   <para>
    <application>pgbench</application> executes test scripts chosen randomly
@@ -1583,7 +1669,7 @@ END;
    The format of the log is:
 
 <synopsis>
-<replaceable>client_id</replaceable> <replaceable>transaction_no</replaceable> <replaceable>time</replaceable> <replaceable>script_no</replaceable> <replaceable>time_epoch</replaceable> <replaceable>time_us</replaceable> <optional> <replaceable>schedule_lag</replaceable> </optional>
+<replaceable>client_id</replaceable> <replaceable>transaction_no</replaceable> <replaceable>time</replaceable> <replaceable>script_no</replaceable> <replaceable>time_epoch</replaceable> <replaceable>time_us</replaceable> <optional> <replaceable>schedule_lag</replaceable> </optional> <optional> <replaceable>retries</replaceable> </optional>
 </synopsis>
 
    where
@@ -1604,6 +1690,19 @@ END;
    When both <option>--rate</option> and <option>--latency-limit</option> are used,
    the <replaceable>time</replaceable> for a skipped transaction will be reported as
    <literal>skipped</literal>.
+   <replaceable>retries</replaceable> is the sum of all the retries after the
+   serialization or deadlock failures during the current script execution. It is
+   only present when the maximum number of tries for transactions is more than 1
+   (<option>--max-tries</option>) and/or the maximum time of tries for
+   transactions is used (<option>--latency-limit</option>). If the transaction
+   ended with an error, its <replaceable>time</replaceable> will be reported as
+   <literal>failed</literal>. If you use the option
+   <option>--errors-detailed</option> the <replaceable>time</replaceable> of the
+   failed transaction will be reported as
+   <literal>serialization_error</literal> / <literal>deadlock_error</literal> /
+   <literal>another_sql_error</literal> / <literal>meta_command_error</literal>
+   depending on the type of error (see <xref linkend="errors-and-retries"
+   endterm="errors-and-retries-title"/> for more information).
   </para>
 
   <para>
@@ -1633,6 +1732,24 @@ END;
   </para>
 
   <para>
+   The following example shows a snippet of a log file with errors and retries,
+   with the maximum number of tries set to 10 (note the additional
+   <replaceable>retries</replaceable> column):
+<screen>
+3 0 47423 0 1499414498 34501 4
+3 1 8333 0 1499414498 42848 1
+3 2 8358 0 1499414498 51219 1
+4 0 72345 0 1499414498 59433 7
+1 3 41718 0 1499414498 67879 5
+1 4 8416 0 1499414498 76311 1
+3 3 33235 0 1499414498 84469 4
+0 0 failed 0 1499414498 84905 10
+2 0 failed 0 1499414498 86248 10
+3 4 8307 0 1499414498 92788 1
+</screen>
+  </para>
+
+  <para>
    When running a long test on hardware that can handle a lot of transactions,
    the log files can become very large.  The <option>--sampling-rate</option> option
    can be used to log only a random sample of transactions.
@@ -1647,7 +1764,7 @@ END;
    format is used for the log files:
 
 <synopsis>
-<replaceable>interval_start</replaceable> <replaceable>num_transactions</replaceable> <replaceable>sum_latency</replaceable> <replaceable>sum_latency_2</replaceable> <replaceable>min_latency</replaceable> <replaceable>max_latency</replaceable> <optional> <replaceable>sum_lag</replaceable> <replaceable>sum_lag_2</replaceable> <replaceable>min_lag</replaceable> <replaceable>max_lag</replaceable> <optional> <replaceable>skipped</replaceable> </optional> </optional>
+<replaceable>interval_start</replaceable> <replaceable>num_transactions</replaceable> <replaceable>sum_latency</replaceable> <replaceable>sum_latency_2</replaceable> <replaceable>min_latency</replaceable> <replaceable>max_latency</replaceable> { <replaceable>failed_tx</replaceable> <optional> | <replaceable>serialization_errors</replaceable> <replaceable>deadlock_errors</replaceable> <replaceable>other_sql_errors</replaceable> <replaceable>meta_command_errors</replaceable> </optional> } <optional> <replaceable>sum_lag</replaceable> <replaceable>sum_lag_2</replaceable> <replaceable>min_lag</replaceable> <replaceable>max_lag</replaceable> <optional> <replaceable>skipped</replaceable> </optional> </optional> <optional> <replaceable>retried_tx</replaceable> <replaceable>retries</replaceable> </optional>
 </synopsis>
 
    where
@@ -1661,7 +1778,20 @@ END;
    transaction latencies within the interval,
    <replaceable>min_latency</replaceable> is the minimum latency within the interval,
    and
-   <replaceable>max_latency</replaceable> is the maximum latency within the interval.
+   <replaceable>max_latency</replaceable> is the maximum latency within the interval,
+   <replaceable>failed_tx</replaceable> is the number of transactions that ended
+   with a failed SQL or meta command within the interval. If you use the option
+   <option>--errors-detailed</option>, instead of the sum of all failed
+   transactions you will get more detailed statistics for the failed
+   transactions grouped by the following types:
+   <replaceable>serialization_errors</replaceable> is the number of transactions
+   that got a serialization failure and were not retried after this,
+   <replaceable>deadlock_errors</replaceable> is the number of transactions that
+   got a deadlock failure and were not retried after this,
+   <replaceable>other_sql_errors</replaceable> is the number of transactions
+   that got a different error in the SQL command,
+   <replaceable>meta_command_errors</replaceable> is the number of transactions
+   that got an error in the meta command.
    The next fields,
    <replaceable>sum_lag</replaceable>, <replaceable>sum_lag_2</replaceable>, <replaceable>min_lag</replaceable>,
    and <replaceable>max_lag</replaceable>, are only present if the <option>--rate</option>
@@ -1669,21 +1799,28 @@ END;
    They provide statistics about the time each transaction had to wait for the
    previous one to finish, i.e. the difference between each transaction's
    scheduled start time and the time it actually started.
-   The very last field, <replaceable>skipped</replaceable>,
+   The next field, <replaceable>skipped</replaceable>,
    is only present if the <option>--latency-limit</option> option is used, too.
    It counts the number of transactions skipped because they would have
    started too late.
+   The <replaceable>retried_tx</replaceable> and
+   <replaceable>retries</replaceable> fields are only present if the maximum
+   number of tries for transactions is more than 1
+   (<option>--max-tries</option>) and/or the maximum time of tries for
+   transactions is used (<option>--latency-limit</option>). They report the
+   number of retried transactions and the sum of all the retries after
+   serialization or deadlock failures within the interval.
    Each transaction is counted in the interval when it was committed.
   </para>
 
   <para>
    Here is some example output:
 <screen>
-1345828501 5601 1542744 483552416 61 2573
-1345828503 7884 1979812 565806736 60 1479
-1345828505 7208 1979422 567277552 59 1391
-1345828507 7685 1980268 569784714 60 1398
-1345828509 7073 1979779 573489941 236 1411
+1345828501 5601 1542744 483552416 61 2573 0
+1345828503 7884 1979812 565806736 60 1479 0
+1345828505 7208 1979422 567277552 59 1391 0
+1345828507 7685 1980268 569784714 60 1398 0
+1345828509 7073 1979779 573489941 236 1411 0
 </screen></para>
 
   <para>
@@ -1695,13 +1832,45 @@ END;
  </refsect2>
 
  <refsect2>
-  <title>Per-Statement Latencies</title>
+  <title>Per-Statement Report</title>
+
+  <para>
+   With the <option>-r</option> option, <application>pgbench</application>
+   collects the following statistics for each statement:
+   <itemizedlist>
+     <listitem>
+       <para>
+         <literal>latency</literal> &mdash; elapsed transaction time for each
+         statement. <application>pgbench</application> reports an average value
+         of all successful runs of the statement.
+       </para>
+     </listitem>
+     <listitem>
+       <para>
+         The number of errors in this statement. See
+         <xref linkend="errors-and-retries" endterm="errors-and-retries-title"/>
+         for more information.
+       </para>
+     </listitem>
+     <listitem>
+       <para>
+         The number of retries after a serialization or a deadlock failure in
+         this statement. See <xref linkend="errors-and-retries"
+         endterm="errors-and-retries-title"/> for more information.
+       </para>
+     </listitem>
+   </itemizedlist>
+  </para>
+
+  <para>
+   The report displays retry statistics only if the maximum number of tries for
+   transactions is more than 1 (<option>--max-tries</option>) and/or the maximum
+   time of tries for transactions is used (<option>--latency-limit</option>).
+  </para>
 
   <para>
-   With the <option>-r</option> option, <application>pgbench</application> collects
-   the elapsed transaction time of each statement executed by every
-   client.  It then reports an average of those values, referred to
-   as the latency for each statement, after the benchmark has finished.
+   All values are computed for each statement executed by every client and are
+   reported after the benchmark has finished.
   </para>
 
   <para>
@@ -1715,27 +1884,69 @@ number of clients: 10
 number of threads: 1
 number of transactions per client: 1000
 number of transactions actually processed: 10000/10000
+maximum number of tries: 1
 latency average = 15.844 ms
 latency stddev = 2.715 ms
 tps = 618.764555 (including connections establishing)
 tps = 622.977698 (excluding connections establishing)
-statement latencies in milliseconds:
-        0.002  \set aid random(1, 100000 * :scale)
-        0.005  \set bid random(1, 1 * :scale)
-        0.002  \set tid random(1, 10 * :scale)
-        0.001  \set delta random(-5000, 5000)
-        0.326  BEGIN;
-        0.603  UPDATE pgbench_accounts SET abalance = abalance + :delta WHERE aid = :aid;
-        0.454  SELECT abalance FROM pgbench_accounts WHERE aid = :aid;
-        5.528  UPDATE pgbench_tellers SET tbalance = tbalance + :delta WHERE tid = :tid;
-        7.335  UPDATE pgbench_branches SET bbalance = bbalance + :delta WHERE bid = :bid;
-        0.371  INSERT INTO pgbench_history (tid, bid, aid, delta, mtime) VALUES (:tid, :bid, :aid, :delta, CURRENT_TIMESTAMP);
-        1.212  END;
+statement latencies in milliseconds and errors:
+        0.002  0  \set aid random(1, 100000 * :scale)
+        0.005  0  \set bid random(1, 1 * :scale)
+        0.002  0  \set tid random(1, 10 * :scale)
+        0.001  0  \set delta random(-5000, 5000)
+        0.326  0  BEGIN;
+        0.603  0  UPDATE pgbench_accounts SET abalance = abalance + :delta WHERE aid = :aid;
+        0.454  0  SELECT abalance FROM pgbench_accounts WHERE aid = :aid;
+        5.528  0  UPDATE pgbench_tellers SET tbalance = tbalance + :delta WHERE tid = :tid;
+        7.335  0  UPDATE pgbench_branches SET bbalance = bbalance + :delta WHERE bid = :bid;
+        0.371  0  INSERT INTO pgbench_history (tid, bid, aid, delta, mtime) VALUES (:tid, :bid, :aid, :delta, CURRENT_TIMESTAMP);
+        1.212  0  END;
+</screen>
+
+   Another example of output for the default script using serializable default
+   transaction isolation level (<command>PGOPTIONS='-c
+   default_transaction_isolation=serializable' pgbench ...</command>):
+<screen>
+starting vacuum...end.
+transaction type: &lt;builtin: TPC-B (sort of)&gt;
+scaling factor: 1
+query mode: simple
+number of clients: 10
+number of threads: 1
+number of transactions per client: 1000
+number of transactions actually processed: 4473/10000
+number of errors: 5527 (55.270%)
+number of serialization errors: 5527 (55.270%)
+number of retried: 7467 (74.670%)
+number of retries: 257244
+maximum number of tries: 100
+number of transactions above the 100.0 ms latency limit: 5766/10000 (57.660 %) (including errors)
+latency average = 41.169 ms
+latency stddev = 51.783 ms
+tps = 50.322494 (including connections establishing)
+tps = 50.324595 (excluding connections establishing)
+statement latencies in milliseconds, errors and retries:
+  0.004     0       0  \set aid random(1, 100000 * :scale)
+  0.000     0       0  \set bid random(1, 1 * :scale)
+  0.000     0       0  \set tid random(1, 10 * :scale)
+  0.000     0       0  \set delta random(-5000, 5000)
+  0.213     0       0  BEGIN;
+  0.393     0       0  UPDATE pgbench_accounts
+                       SET abalance = abalance + :delta WHERE aid = :aid;
+  0.332     0       0  SELECT abalance FROM pgbench_accounts WHERE aid = :aid;
+  0.409  4971  250265  UPDATE pgbench_tellers
+                       SET tbalance = tbalance + :delta WHERE tid = :tid;
+  0.311   556    6975  UPDATE pgbench_branches
+                       SET bbalance = bbalance + :delta WHERE bid = :bid;
+  0.299     0       0  INSERT INTO pgbench_history
+                              (tid, bid, aid, delta, mtime)
+                       VALUES (:tid, :bid, :aid, :delta, CURRENT_TIMESTAMP);
+  0.520     0       4  END;
 </screen>
   </para>
 
   <para>
-   If multiple script files are specified, the averages are reported
+   If multiple script files are specified, all statistics are reported
    separately for each script file.
   </para>
 
@@ -1749,6 +1960,136 @@ statement latencies in milliseconds:
   </para>
  </refsect2>
 
+ <refsect2 id="errors-and-retries">
+  <title id="errors-and-retries-title">Errors and Serialization/Deadlock Retries</title>
+
+  <para>
+   When executing <application>pgbench</application>, there're three main types of
+   errors:
+   <itemizedlist>
+     <listitem>
+       <para>
+         Errors of the main program. They are the most serious and always result
+         in an immediate exit from the <application>pgbench</application> with
+         the corresponding error message. They include:
+         <itemizedlist>
+           <listitem>
+             <para>
+               errors at the beginning of the <application>pgbench</application>
+               (e.g. an invalid option value);
+             </para>
+           </listitem>
+           <listitem>
+             <para>
+               errors in the initialization mode (e.g. the query to create
+               tables for built-in scripts fails);
+             </para>
+           </listitem>
+           <listitem>
+             <para>
+               errors before starting threads (e.g. we could not connect to the
+               database server / the syntax error in the meta command / thread
+               creation failure);
+             </para>
+           </listitem>
+           <listitem>
+             <para>
+               internal <application>pgbench</application> errors (which are
+               supposed to never occur..).
+             </para>
+           </listitem>
+         </itemizedlist>
+       </para>
+     </listitem>
+     <listitem>
+       <para>
+         Errors when the thread manages its clients (e.g. the client could not
+         start a connection to the database server / the socket for connecting
+         the client to the database server has become invalid). In such cases
+         all clients of this thread stop while other threads continue to work.
+       </para>
+     </listitem>
+     <listitem>
+       <para>
+         Directly client errors. They lead to immediate exit from the
+         <application>pgbench</application> with the corresponding error message
+         only in the case of an internal <application>pgbench</application>
+         error (which are supposed to never occur..). Otherwise in the worst
+         case they only lead to the abortion of the failed client while other
+         clients continue their run (but most client errors are handled without
+         an abortion of the client and reported separately, see below). Later in
+         this section it is assumed that the discussed failures and errors are
+         only the directly client errors and they are not internal
+         <application>pgbench</application> errors.
+       </para>
+     </listitem>
+   </itemizedlist>
+  </para>
+
+  <para>
+   Client's run is aborted only in case of a serious error, for example, the
+   connection with the database server was lost. Otherwise if the execution of
+   SQL or meta command fails, the failed transaction is always rolled back which
+   also includes setting the client variables as they were before the run of
+   this transaction (it is assumed that one transaction script contains only one
+   transaction; see <xref linkend="transactions-and-scripts"
+   endterm="transactions-and-scripts-title"/> for more information).
+   Transactions with serialization or deadlock failures are repeated after
+   rollbacks until they complete successfully or reach the maximum number of
+   tries (specified by the <option>--max-tries</option> option) / the maximum
+   time of tries (specified by the <option>--latency-limit</option> option). If
+   the last transaction run fails, this transaction will be reported as failed.
+  </para>
+
+  <note>
+   <para>
+    Be careful when repeating scripts that contain multiple transactions: the
+    script is always retried completely, so the successful transactions can be
+    performed several times.
+   </para>
+   <para>
+    Be careful when repeating scripts that continue the transaction block from
+    the previous script: the command used to start this transaction block like
+    all the commands from previous scripts is never used on a retry.
+   </para>
+   <para>
+    Be careful when repeating transactions with shell commands. Unlike the
+    results of SQL commands, the results of shell commands are not rolled back,
+    except for the variable value of the <command>\setshell</command> command.
+   </para>
+  </note>
+
+  <para>
+   The latency of a successful transaction includes the entire time of
+   transaction execution with rollbacks and retries. The latency for failed
+   transactions and commands is not computed separately.
+  </para>
+
+  <para>
+   The main report contains the number of failed transactions if it is non-zero
+   (to group them into basic types, use the option
+   <option>--errors-detailed</option>). If the total number of retried
+   transactions is non-zero, the main report also contains the statistics
+   related to retries: the total number of retried transactions and total number
+   of retries. The per-script report inherits all these fields from the main
+   report. The per-statement report displays retry statistics only if the
+   maximum number of tries for transactions is more than 1
+   (<option>--max-tries</option>) and/or the maximum time of tries for
+   transactions is used (<option>--latency-limit</option>). A retry is reported
+   for the command where the failure occured during the current script
+   execution.
+  </para>
+
+  <para>
+   If you want to distinguish between failures or errors by type (including
+   which limit for retries was violated and how far it was exceeded for the
+   serialization/deadlock errors), use the <application>pgbench</application>
+   debugging output created with the option <option>--debug-fails</option> or
+   <option>--debug</option>. The first option is recommended for this purpose
+   because with the second option the debugging output can be very large.
+  </para>
+ </refsect2>
+
  <refsect2>
   <title>Good Practices</title>
 
diff --git a/src/bin/pgbench/pgbench.c b/src/bin/pgbench/pgbench.c
index 4f8700b..1ed24fd 100644
--- a/src/bin/pgbench/pgbench.c
+++ b/src/bin/pgbench/pgbench.c
@@ -59,6 +59,8 @@
 
 #include "pgbench.h"
 
+#define ERRCODE_T_R_SERIALIZATION_FAILURE  "40001"
+#define ERRCODE_T_R_DEADLOCK_DETECTED  "40P01"
 #define ERRCODE_UNDEFINED_TABLE  "42P01"
 
 /*
@@ -187,9 +189,30 @@ bool		progress_timestamp = false; /* progress report with Unix time */
 int			nclients = 1;		/* number of clients */
 int			nthreads = 1;		/* number of threads */
 bool		is_connect;			/* establish connection for each transaction */
-bool		is_latencies;		/* report per-command latencies */
+bool		report_per_command = false;	/* report per-command latencies, retries
+										 * after the failures and errors
+										 * (failures without retrying) */
 int			main_pid;			/* main process id used in log filename */
 
+/*
+ * There're different types of restrictions for deciding that the current failed
+ * transaction can no longer be retried and should be reported as failed:
+ * - max_tries can be used to limit the number of tries;
+ * - latency_limit can be used to limit the total time of tries.
+ *
+ * They can be combined together, and you need to use at least one of them to
+ * retry the failed transactions. If none of them is used, the default value of
+ * max_tries is set to 1 and failed transactions are not retried at all.
+ */
+uint32		max_tries = 0;		/* we cannot retry a failed transaction if its
+								 * number of tries reaches this maximum; if its
+								 * value is zero, it is not used */
+
+#define RETRIES_ENABLED			(max_tries > 1 || latency_limit)
+
+bool		errors_detailed = false;	/* whether to group errors in reports or
+										 * logs by basic types */
+
 char	   *pghost = "";
 char	   *pgport = "";
 char	   *login = NULL;
@@ -267,9 +290,58 @@ typedef struct SimpleStats
 typedef struct StatsData
 {
 	time_t		start_time;		/* interval start time, for aggregates */
-	int64		cnt;			/* number of transactions, including skipped */
+
+	/*
+	 * All the transactions fall into 2 main types: is there any command that
+	 * got a failure during the last execution of the transaction script or not?
+	 * Thus
+	 *
+	 * the number of all transactions =
+	 *   cnt (the number of successful ones) +
+	 *   ecnt (the number of failed transactions).
+	 *
+	 * A successful transaction can be one of several types:
+	 *
+	 * cnt (the number of successful transactions) =
+	 *   skipped (it was to late to execute them) +
+	 *   retried (they got a serialization or a deadlock failure(s), but were
+	 *            successfully retried from the very beginning) +
+	 *   number of other transactions.
+	 *
+	 * A failed transaction can be one of several types:
+	 *
+	 * ecnt (the number of failed transactions) =
+	 *   serialization_errors (they got a serialization failure and were not
+	 *                         retried) +
+	 *   deadlock_errors (they got a deadlock failure and were not retried) +
+	 *   other_sql_errors (they got a different error in the SQL command) +
+	 *   meta_command_errors (they got an error in the meta command).
+	 *
+	 * If the transaction was retried after a serialization or a deadlock
+	 * failure this does not guarantee that this retry was successfull. Thus
+	 *
+	 * number of retries =
+	 *   number of retries in (retried + serialization_errors + deadlock_errors)
+	 *   transactions.
+	 */
+	int64		cnt;			/* number of sucessfull transactions, including
+								 * skipped */
 	int64		skipped;		/* number of transactions skipped under --rate
 								 * and --latency-limit */
+	int64		retries;		/* number of retries after a serialization or a
+								 * deadlock failure in all the transactions */
+	int64		retried;		/* number of transactions that were retried
+								 * after a serialization or a deadlock
+								 * failure */
+	int64		serialization_errors;	/* number of transactions that were not
+										 * retried after a serialization
+										 * failure */
+	int64		deadlock_errors;	/* number of transactions that were not
+									 * retried after a deadlock failure */
+	int64		other_sql_errors;	/* number of transactions with a different
+									 * error in the SQL command */
+	int64		meta_command_errors;	/* number of transactions with an error
+										 * in the meta command */
 	SimpleStats latency;
 	SimpleStats lag;
 } StatsData;
@@ -283,6 +355,29 @@ typedef struct
 } RandomState;
 
 /*
+ * Data structure for repeating a transaction from the beginnning with the same
+ * parameters.
+ */
+typedef struct
+{
+	RandomState random_state;	/* random seed */
+	Variables   variables;		/* client variables */
+} RetryState;
+
+/*
+ * For the failures during script execution.
+ */
+typedef enum FailureStatus
+{
+	NO_FAILURE = 0,
+	META_COMMAND_FAILURE,
+	ANOTHER_SQL_FAILURE,		/* other failures in SQL commands that are not
+								 * listed by themselves below */
+	SERIALIZATION_FAILURE,
+	DEADLOCK_FAILURE
+} FailureStatus;
+
+/*
  * Connection state machine states.
  */
 typedef enum
@@ -337,6 +432,23 @@ typedef enum
 	CSTATE_END_COMMAND,
 
 	/*
+	 * States for failed commands.
+	 *
+	 * If the SQL/meta command fails, in CSTATE_FAILURE abort the current
+	 * transaction execution, roll back the failed transaction block if any and
+	 * clear the conditional stack.  Then go to CSTATE_RETRY.  If this is a
+	 * serialization or deadlock failure and we can re-execute the transaction
+	 * from the very beginning, report this as a failure, set the same
+	 * parameters for the transaction execution as in the previous tries and
+	 * process the first transaction command in CSTATE_START_COMMAND. Otherwise,
+	 * report this as an error, set the parameters for the transaction execution
+	 * as they were before the first run of this transaction (except for a
+	 * random state) and go to CSTATE_END_TX to complete this transaction.
+	 */
+	CSTATE_FAILURE,
+	CSTATE_RETRY,
+
+	/*
 	 * CSTATE_END_TX performs end-of-transaction processing.  Calculates
 	 * latency, and logs the transaction.  In --connect mode, closes the
 	 * current connection.  Chooses the next script to execute and starts over
@@ -383,6 +495,18 @@ typedef struct
 
 	bool		prepared[MAX_SCRIPTS];	/* whether client prepared the script */
 
+	/*
+	 * For processing errors and repeating transactions with serialization or
+	 * deadlock failures:
+	 */
+	FailureStatus failure_status;	/* the failure status of the current
+									 * transaction execution; this is NO_FAILURE
+									 * if there were no failures */
+	RetryState  retry_state;
+	uint32			retries;	/* how many times have we already retried the
+								 * current transaction after a serialization or
+								 * a deadlock failure? */
+
 	/* per client collected stats */
 	int64		cnt;			/* client transaction count, for -t */
 	int			ecnt;			/* error count */
@@ -445,7 +569,8 @@ typedef struct
 	instr_time	start_time;		/* thread start time */
 	instr_time	conn_time;
 	StatsData	stats;
-	int64		latency_late;	/* executed but late transactions */
+	int64		latency_late;	/* executed but late transactions (including
+								 * errors) */
 } TState;
 
 #define INVALID_THREAD		((pthread_t) 0)
@@ -491,6 +616,9 @@ typedef struct
 	char	   *argv[MAX_ARGS]; /* command word list */
 	PgBenchExpr *expr;			/* parsed expression, if needed */
 	SimpleStats stats;			/* time spent in this command */
+	int64		retries;		/* number of retries after a serialization or a
+								 * deadlock failure */
+	int64		ecnt;			/* number of failures that were not retried */
 } Command;
 
 typedef struct ParsedScript
@@ -560,9 +688,15 @@ typedef enum ErrorLevel
 	DEBUG,
 
 	/*
+	 * Normal failure of the SQL/meta command, or processing of the failed
+	 * transaction (its end/retry).
+	 */
+	DEBUG_FAIL,
+
+	/*
 	 * To report:
-	 * - abortion of the client (something bad e.g. the SQL/meta command failed
-	 *   or the connection with the backend was lost);
+	 * - abortion of the client (something serious e.g. connection with the
+	 *   backend was lost);
 	 * - the log messages of the main program;
 	 * - PGBENCH_DEBUG messages.
 	 */
@@ -596,7 +730,6 @@ static void finishCon(CState *st);
 static void pgbench_error(ErrorLevel elevel,
 						  const char *fmt,...) pg_attribute_printf(2, 3);
 
-
 /* callback functions for our flex lexer */
 static const PsqlScanCallbacks pgbench_callbacks = {
 	NULL,						/* don't need get_variable functionality */
@@ -643,15 +776,18 @@ usage(void)
 		   "                           protocol for submitting queries (default: simple)\n"
 		   "  -n, --no-vacuum          do not run VACUUM before tests\n"
 		   "  -P, --progress=NUM       show thread progress report every NUM seconds\n"
-		   "  -r, --report-latencies   report average latency per command\n"
+		   "  -r, --report-per-command report latencies, errors and retries per command\n"
 		   "  -R, --rate=NUM           target rate in transactions per second\n"
 		   "  -s, --scale=NUM          report this scale factor in output\n"
 		   "  -t, --transactions=NUM   number of transactions each client runs (default: 10)\n"
 		   "  -T, --time=NUM           duration of benchmark test in seconds\n"
 		   "  -v, --vacuum-all         vacuum all four standard tables before tests\n"
 		   "  --aggregate-interval=NUM aggregate data over NUM seconds\n"
+		   "  --debug-fails            print debugging output only for failures and errors\n"
+		   "  --errors-detailed        report the errors grouped by basic types\n"
 		   "  --log-prefix=PREFIX      prefix for transaction time log file\n"
 		   "                           (default: \"pgbench_log\")\n"
+		   "  --max-tries=NUM          max number of tries to run transaction\n"
 		   "  --progress-timestamp     use Unix epoch timestamps for progress\n"
 		   "  --random-seed=SEED       set random seed (\"time\", \"rand\", integer)\n"
 		   "  --sampling-rate=NUM      fraction of transactions to log (e.g., 0.01 for 1%%)\n"
@@ -1112,6 +1248,12 @@ initStats(StatsData *sd, time_t start_time)
 	sd->start_time = start_time;
 	sd->cnt = 0;
 	sd->skipped = 0;
+	sd->retries = 0;
+	sd->retried = 0;
+	sd->serialization_errors = 0;
+	sd->deadlock_errors = 0;
+	sd->other_sql_errors = 0;
+	sd->meta_command_errors = 0;
 	initSimpleStats(&sd->latency);
 	initSimpleStats(&sd->lag);
 }
@@ -1120,10 +1262,47 @@ initStats(StatsData *sd, time_t start_time)
  * Accumulate one additional item into the given stats object.
  */
 static void
-accumStats(StatsData *stats, bool skipped, double lat, double lag)
+accumStats(StatsData *stats, bool skipped, double lat, double lag,
+		   FailureStatus failure_status, int64 retries)
 {
-	stats->cnt++;
+	/*
+	 * Record the number of retries regardless of whether the transaction was
+	 * successful or failed.
+	 */
+	stats->retries += retries;
+	if (retries > 0)
+		stats->retried++;
+
+	/* Record the failed transaction */
+	if (failure_status != NO_FAILURE)
+	{
+		if (failure_status == SERIALIZATION_FAILURE)
+		{
+			stats->serialization_errors++;
+		}
+		else if (failure_status == DEADLOCK_FAILURE)
+		{
+			stats->deadlock_errors++;
+		}
+		else if (failure_status == ANOTHER_SQL_FAILURE)
+		{
+			stats->other_sql_errors++;
+		}
+		else if (failure_status == META_COMMAND_FAILURE)
+		{
+			stats->meta_command_errors++;
+		}
+		else
+		{
+			/* internal error which should never occur */
+			pgbench_error(ERROR, "unexpected failure status: %d\n",
+						  failure_status);
+		}
+		return;
+	}
 
+	/* Record the successful transaction */
+	stats->cnt++;
 	if (skipped)
 	{
 		/* no latency to record on skipped transactions */
@@ -1376,7 +1555,7 @@ makeVariableValue(Variable *var)
 
 		if (sscanf(var->svalue, "%lf%c", &dv, &xs) != 1)
 		{
-			pgbench_error(LOG,
+			pgbench_error(DEBUG_FAIL,
 						  "malformed variable \"%s\" value: \"%s\"\n",
 						  var->name, var->svalue);
 			return false;
@@ -1505,7 +1684,7 @@ lookupCreateVariable(Variables *variables, const char *context, char *name,
 	 * About the error level used: if we process client commands, it a normal
 	 * failure; otherwise it is not and we exit the program.
 	 */
-	ErrorLevel elevel = client ? LOG : ERROR;
+	ErrorLevel elevel = client ? DEBUG_FAIL : ERROR;
 
 	var = lookupVariable(variables, name);
 	if (var == NULL)
@@ -1726,11 +1905,12 @@ coerceToBool(PgBenchValue *pval, bool *bval)
 	}
 	else						/* NULL, INT or DOUBLE */
 	{
-		/* we are sure that the function valueTypeName only is always called */
-		Assert(LOG >= log_min_messages);
-
-		pgbench_error(LOG, "cannot coerce %s to boolean\n",
-					  valueTypeName(pval));
+		/* call the function valueTypeName only if necessary */
+		if (DEBUG_FAIL >= log_min_messages)
+		{
+			pgbench_error(DEBUG_FAIL, "cannot coerce %s to boolean\n",
+						  valueTypeName(pval));
+		}
 		*bval = false;			/* suppress uninitialized-variable warnings */
 		return false;
 	}
@@ -1775,7 +1955,7 @@ coerceToInt(PgBenchValue *pval, int64 *ival)
 
 		if (dval < PG_INT64_MIN || PG_INT64_MAX < dval)
 		{
-			pgbench_error(LOG, "double to int overflow for %f\n", dval);
+			pgbench_error(DEBUG_FAIL, "double to int overflow for %f\n", dval);
 			return false;
 		}
 		*ival = (int64) dval;
@@ -1783,10 +1963,12 @@ coerceToInt(PgBenchValue *pval, int64 *ival)
 	}
 	else						/* BOOLEAN or NULL */
 	{
-		/* we are sure that the function valueTypeName is always called */
-		Assert(LOG >= log_min_messages);
-
-		pgbench_error(LOG, "cannot coerce %s to int\n", valueTypeName(pval));
+		/* call the function valueTypeName only if necessary */
+		if (DEBUG_FAIL >= log_min_messages)
+		{
+			pgbench_error(DEBUG_FAIL, "cannot coerce %s to int\n",
+						  valueTypeName(pval));
+		}
 		return false;
 	}
 }
@@ -1807,10 +1989,12 @@ coerceToDouble(PgBenchValue *pval, double *dval)
 	}
 	else						/* BOOLEAN or NULL */
 	{
-		/* we are sure that the function valueTypeName is always called */
-		Assert(LOG >= log_min_messages);
-
-		pgbench_error(LOG, "cannot coerce %s to double\n", valueTypeName(pval));
+		/* call the function valueTypeName only if necessary */
+		if (DEBUG_FAIL >= log_min_messages)
+		{
+			pgbench_error(DEBUG_FAIL, "cannot coerce %s to double\n",
+						  valueTypeName(pval));
+		}
 		return false;
 	}
 }
@@ -1991,7 +2175,8 @@ evalStandardFunc(TState *thread, CState *st,
 
 	if (l != NULL)
 	{
-		pgbench_error(LOG, "too many function arguments, maximum is %d\n",
+		pgbench_error(DEBUG_FAIL,
+					  "too many function arguments, maximum is %d\n",
 					  MAX_FARGS);
 		return false;
 	}
@@ -2115,7 +2300,7 @@ evalStandardFunc(TState *thread, CState *st,
 						case PGBENCH_MOD:
 							if (ri == 0)
 							{
-								pgbench_error(LOG, "division by zero\n");
+								pgbench_error(DEBUG_FAIL, "division by zero\n");
 								return false;
 							}
 							/* special handling of -1 divisor */
@@ -2126,7 +2311,7 @@ evalStandardFunc(TState *thread, CState *st,
 									/* overflow check (needed for INT64_MIN) */
 									if (li == PG_INT64_MIN)
 									{
-										pgbench_error(LOG,
+										pgbench_error(DEBUG_FAIL,
 													  "bigint out of range\n");
 										return false;
 									}
@@ -2393,13 +2578,13 @@ evalStandardFunc(TState *thread, CState *st,
 				/* check random range */
 				if (imin > imax)
 				{
-					pgbench_error(LOG, "empty range given to random\n");
+					pgbench_error(DEBUG_FAIL, "empty range given to random\n");
 					return false;
 				}
 				else if (imax - imin < 0 || (imax - imin) + 1 < 0)
 				{
 					/* prevent int overflows in random functions */
-					pgbench_error(LOG, "random range is too large\n");
+					pgbench_error(DEBUG_FAIL, "random range is too large\n");
 					return false;
 				}
 
@@ -2421,7 +2606,7 @@ evalStandardFunc(TState *thread, CState *st,
 					{
 						if (param < MIN_GAUSSIAN_PARAM)
 						{
-							pgbench_error(LOG,
+							pgbench_error(DEBUG_FAIL,
 										  "gaussian parameter must be at least %f (not %f)\n",
 										  MIN_GAUSSIAN_PARAM, param);
 							return false;
@@ -2435,7 +2620,7 @@ evalStandardFunc(TState *thread, CState *st,
 					{
 						if (param <= 0.0 || param == 1.0 || param > MAX_ZIPFIAN_PARAM)
 						{
-							pgbench_error(LOG,
+							pgbench_error(DEBUG_FAIL,
 										  "zipfian parameter must be in range (0, 1) U (1, %d] (got %f)\n",
 										  MAX_ZIPFIAN_PARAM, param);
 							return false;
@@ -2449,7 +2634,7 @@ evalStandardFunc(TState *thread, CState *st,
 					{
 						if (param <= 0.0)
 						{
-							pgbench_error(LOG,
+							pgbench_error(DEBUG_FAIL,
 										  "exponential parameter must be greater than zero (got %f)\n",
 										  param);
 							return false;
@@ -2562,7 +2747,7 @@ evaluateExpr(TState *thread, CState *st, PgBenchExpr *expr, PgBenchValue *retval
 
 				if ((var = lookupVariable(&st->variables, expr->u.variable.varname)) == NULL)
 				{
-					pgbench_error(LOG, "undefined variable \"%s\"\n",
+					pgbench_error(DEBUG_FAIL, "undefined variable \"%s\"\n",
 								  expr->u.variable.varname);
 					return false;
 				}
@@ -2662,7 +2847,7 @@ runShellCommand(Variables *variables, char *variable, char **argv, int argc)
 		}
 		else if ((arg = getVariable(variables, argv[i] + 1)) == NULL)
 		{
-			pgbench_error(LOG, "%s: undefined variable \"%s\"\n",
+			pgbench_error(DEBUG_FAIL, "%s: undefined variable \"%s\"\n",
 						  argv[0], argv[i]);
 			return false;
 		}
@@ -2670,7 +2855,8 @@ runShellCommand(Variables *variables, char *variable, char **argv, int argc)
 		arglen = strlen(arg);
 		if (len + arglen + (i > 0 ? 1 : 0) >= SHELL_COMMAND_SIZE - 1)
 		{
-			pgbench_error(LOG, "%s: shell command is too long\n", argv[0]);
+			pgbench_error(DEBUG_FAIL, "%s: shell command is too long\n",
+						  argv[0]);
 			return false;
 		}
 
@@ -2689,8 +2875,8 @@ runShellCommand(Variables *variables, char *variable, char **argv, int argc)
 		{
 			if (!timer_exceeded)
 			{
-				pgbench_error(LOG, "%s: could not launch shell command\n",
-							  argv[0]);
+				pgbench_error(DEBUG_FAIL,
+							  "%s: could not launch shell command\n", argv[0]);
 			}
 			return false;
 		}
@@ -2700,14 +2886,16 @@ runShellCommand(Variables *variables, char *variable, char **argv, int argc)
 	/* Execute the command with pipe and read the standard output. */
 	if ((fp = popen(command, "r")) == NULL)
 	{
-		pgbench_error(LOG, "%s: could not launch shell command\n", argv[0]);
+		pgbench_error(DEBUG_FAIL, "%s: could not launch shell command\n",
+					  argv[0]);
 		return false;
 	}
 	if (fgets(res, sizeof(res), fp) == NULL)
 	{
 		if (!timer_exceeded)
 		{
-			pgbench_error(LOG, "%s: could not read result of shell command\n",
+			pgbench_error(DEBUG_FAIL,
+						  "%s: could not read result of shell command\n",
 						  argv[0]);
 		}
 		(void) pclose(fp);
@@ -2715,7 +2903,8 @@ runShellCommand(Variables *variables, char *variable, char **argv, int argc)
 	}
 	if (pclose(fp) < 0)
 	{
-		pgbench_error(LOG, "%s: could not close shell command\n", argv[0]);
+		pgbench_error(DEBUG_FAIL, "%s: could not close shell command\n",
+					  argv[0]);
 		return false;
 	}
 
@@ -2725,7 +2914,7 @@ runShellCommand(Variables *variables, char *variable, char **argv, int argc)
 		endptr++;
 	if (*res == '\0' || *endptr != '\0')
 	{
-		pgbench_error(LOG,
+		pgbench_error(DEBUG_FAIL,
 					  "%s: shell command must return an integer (not \"%s\")\n",
 					  argv[0], res);
 		return false;
@@ -2747,11 +2936,20 @@ preparedStatementName(char *buffer, int file, int state)
 }
 
 static void
-commandFailed(CState *st, const char *cmd, const char *message)
+commandFailed(CState *st, const char *cmd, const char *message, bool aborted)
 {
-	pgbench_error(LOG,
-				  "client %d aborted in command %d (%s) of script %d; %s\n",
-				  st->id, st->command, cmd, st->use_file, message);
+	if (aborted)
+	{
+		pgbench_error(LOG,
+					  "client %d aborted in command %d (%s) of script %d; %s\n",
+					  st->id, st->command, cmd, st->use_file, message);
+	}
+	else
+	{
+		pgbench_error(DEBUG_FAIL,
+					  "client %d got a failure in command %d (%s) of script %d; %s\n",
+					  st->id, st->command, cmd, st->use_file, message);
+	}
 }
 
 /* return a script number with a weighted choice. */
@@ -2850,7 +3048,6 @@ sendCommand(CState *st, Command *command)
 	{
 		pgbench_error(DEBUG, "client %d could not send %s\n",
 					  st->id, command->argv[0]);
-		st->ecnt++;
 		return false;
 	}
 	else
@@ -2871,7 +3068,7 @@ evaluateSleep(Variables *variables, int argc, char **argv, int *usecs)
 	{
 		if ((var = getVariable(variables, argv[1] + 1)) == NULL)
 		{
-			pgbench_error(LOG, "%s: undefined variable \"%s\"\n",
+			pgbench_error(DEBUG_FAIL, "%s: undefined variable \"%s\"\n",
 						  argv[0], argv[1]);
 			return false;
 		}
@@ -2895,6 +3092,225 @@ evaluateSleep(Variables *variables, int argc, char **argv, int *usecs)
 }
 
 /*
+ * Get the number of all processed transactions including skipped ones and
+ * errors.
+ */
+static int64
+getTotalCnt(const CState *st)
+{
+	return st->cnt + st->ecnt;
+}
+
+static int64
+getEcnt(const StatsData *stats)
+{
+	return (stats->serialization_errors +
+			stats->deadlock_errors +
+			stats->other_sql_errors +
+			stats->meta_command_errors);
+}
+
+/*
+ * Clear the variables in the array. The array itself is not freed.
+ */
+static void
+clearVariables(Variables *variables)
+{
+	Variable   *vars,
+			   *var;
+	int			nvars;
+
+	if (!variables)
+		return;					/* nothing to do here */
+
+	vars = variables->vars;
+	nvars = variables->nvars;
+	for (var = vars; var - vars < nvars; ++var)
+	{
+		pg_free(var->name);
+		pg_free(var->svalue);
+	}
+
+	variables->nvars = 0;
+}
+
+/*
+ * Make a deep copy of variables array.
+ * Before copying the function frees the string fields of the destination
+ * variables and if necessary enlarges their array.
+ * On failure (too many variables are used in the source) returns false.
+ */
+static bool
+copyVariables(Variables *dest, const Variables *source)
+{
+	Variable   *dest_var;
+	const Variable *source_vars,
+				   *source_var;
+	int			nvars;
+
+	if (!dest || !source || dest == source)
+		return true;			/* nothing to do here */
+
+	source_vars = source->vars;
+	nvars = source->nvars;
+
+	/*
+	 * Clear the original variables and make sure that we have enough space for
+	 * the new variables.
+	 */
+
+	clearVariables(dest);
+
+	/*
+	 * In case of an error the client will be aborted so always print an error
+	 * message.
+	 */
+	if (!enlargeVariables(dest, NULL, nvars, LOG))
+		return false;
+
+	/* Make a deep copy of variables array */
+	for (source_var = source_vars, dest_var = dest->vars;
+		 source_var - source_vars < nvars;
+		 ++source_var, ++dest_var)
+	{
+		dest_var->name = pg_strdup(source_var->name);
+		if (source_var->svalue == NULL)
+			dest_var->svalue = NULL;
+		else
+			dest_var->svalue = pg_strdup(source_var->svalue);
+		dest_var->value = source_var->value;
+	}
+	dest->nvars = nvars;
+	dest->vars_sorted = source->vars_sorted;
+	return true;
+}
+
+/*
+ * Returns true if this type of failure can be retried.
+ */
+static bool
+canRetryFailure(FailureStatus failure_status)
+{
+	return (failure_status == SERIALIZATION_FAILURE ||
+			failure_status == DEADLOCK_FAILURE);
+}
+
+/*
+ * Returns true if the failure can be retried.
+ */
+static bool
+canRetry(CState *st, instr_time *now)
+{
+	FailureStatus failure_status = st->failure_status;
+
+	Assert(failure_status != NO_FAILURE);
+
+	/* We can only retry serialization or deadlock failures. */
+	if (!canRetryFailure(failure_status))
+		return false;
+
+	/*
+	 * We must have at least one option to limit the retrying of failed
+	 * transactions.
+	 */
+	Assert(max_tries || latency_limit);
+
+	/*
+	 * We cannot retry the failure if we have reached the maximum number of
+	 * tries.
+	 */
+	if (max_tries && st->retries + 1 >= max_tries)
+		return false;
+
+	/*
+	 * We cannot retry the failure if we spent too much time on this
+	 * transaction.
+	 */
+	if (latency_limit)
+	{
+		if (INSTR_TIME_IS_ZERO(*now))
+			INSTR_TIME_SET_CURRENT(*now);
+
+		if (INSTR_TIME_GET_MICROSEC(*now) - st->txn_scheduled >= latency_limit)
+			return false;
+	}
+
+	/* OK */
+	return true;
+}
+
+/*
+ * Get the failure status from the error code.
+ */
+static FailureStatus
+getSQLFailureStatus(char *sqlState)
+{
+	if (sqlState)
+	{
+		if (strcmp(sqlState, ERRCODE_T_R_SERIALIZATION_FAILURE) == 0)
+			return SERIALIZATION_FAILURE;
+		else if (strcmp(sqlState, ERRCODE_T_R_DEADLOCK_DETECTED) == 0)
+			return DEADLOCK_FAILURE;
+	}
+
+	return ANOTHER_SQL_FAILURE;
+}
+
+static bool
+inTransactionBlock(CState *st)
+{
+	PGTransactionStatusType tx_status;
+
+	tx_status = PQtransactionStatus(st->con);
+	switch (tx_status)
+	{
+		case PQTRANS_IDLE:
+			return false;
+		case PQTRANS_INTRANS:
+		case PQTRANS_INERROR:
+			return true;
+		case PQTRANS_UNKNOWN:
+			/* PQTRANS_UNKNOWN is expected given a broken connection */
+			if (PQstatus(st->con) == CONNECTION_BAD)
+			{		/* there's something wrong */
+				pgbench_error(LOG,
+							  "client %d aborted while receiving the transaction status; perhaps the backend died while processing\n",
+							  st->id);
+				st->state = CSTATE_ABORTED;
+				return false;	/* return value does not matter */
+			}
+		case PQTRANS_ACTIVE:
+		default:
+			/*
+			 * We cannot find out whether we are in a transaction block or not.
+			 * Internal error which should never occur.
+			 */
+			pgbench_error(LOG,
+						  "client %d aborted while receiving the transaction status; unexpected transaction status %d\n",
+						  st->id, tx_status);
+			st->state = CSTATE_ABORTED;
+			return false;		/* return value does not matter */
+	}
+}
+
+/*
+ * If the latency limit is used, return a percentage of the current transaction
+ * latency from the latency limit. Otherwise return zero.
+ */
+static double
+getLatencyUsed(CState *st, instr_time *now)
+{
+	if (!latency_limit)
+		return 0.0;
+
+	if (INSTR_TIME_IS_ZERO(*now))
+		INSTR_TIME_SET_CURRENT(*now);
+
+	return (100.0 * (INSTR_TIME_GET_MICROSEC(*now) - st->txn_scheduled) /
+			latency_limit);
+}
+
+/*
  * Advance the state machine of a connection, if possible.
  */
 static void
@@ -2944,6 +3360,11 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 					st->state = CSTATE_START_TX;
 				/* check consistency */
 				Assert(conditional_stack_empty(st->cstack));
+
+				/* reset transaction variables to default values */
+				st->failure_status = NO_FAILURE;
+				st->retries = 0;
+
 				break;
 
 				/*
@@ -2991,7 +3412,7 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 						INSTR_TIME_SET_CURRENT(now);
 					now_us = INSTR_TIME_GET_MICROSEC(now);
 					while (thread->throttle_trigger < now_us - latency_limit &&
-						   (nxacts <= 0 || st->cnt < nxacts))
+						   (nxacts <= 0 || getTotalCnt(st) < nxacts))
 					{
 						processXactStats(thread, st, &now, true, agg);
 						/* next rendez-vous */
@@ -3001,7 +3422,7 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 						st->txn_scheduled = thread->throttle_trigger;
 					}
 					/* stop client if -t exceeded */
-					if (nxacts > 0 && st->cnt >= nxacts)
+					if (nxacts > 0 && getTotalCnt(st) >= nxacts)
 					{
 						st->state = CSTATE_FINISHED;
 						break;
@@ -3057,6 +3478,22 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 				}
 
 				/*
+				 * It is the first try to run this transaction. Remember its
+				 * parameters just in case if it fails or we should repeat it in
+				 * future.
+				 */
+				memcpy(&(st->retry_state.random_state), &(st->random_state),
+					   sizeof(RandomState));
+				if (!copyVariables(&st->retry_state.variables, &st->variables))
+				{
+					pgbench_error(LOG,
+								  "client %d aborted when preparing to execute a transaction\n",
+								  st->id);
+					st->state = CSTATE_ABORTED;
+					break;
+				}
+
+				/*
 				 * Record transaction start time under logging, progress or
 				 * throttling.
 				 */
@@ -3100,7 +3537,7 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 				 * Record statement start time if per-command latencies are
 				 * requested
 				 */
-				if (is_latencies)
+				if (report_per_command)
 				{
 					if (INSTR_TIME_IS_ZERO(now))
 						INSTR_TIME_SET_CURRENT(now);
@@ -3111,7 +3548,8 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 				{
 					if (!sendCommand(st, command))
 					{
-						commandFailed(st, "SQL", "SQL command send failed");
+						commandFailed(st, "SQL", "SQL command send failed",
+									  true);
 						st->state = CSTATE_ABORTED;
 					}
 					else
@@ -3151,8 +3589,11 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 
 						if (!evaluateSleep(&st->variables, argc, argv, &usec))
 						{
-							commandFailed(st, "sleep", "execution of meta-command failed");
-							st->state = CSTATE_ABORTED;
+							commandFailed(st, "sleep",
+										  "execution of meta-command failed",
+										  false);
+							st->failure_status = META_COMMAND_FAILURE;
+							st->state = CSTATE_FAILURE;
 							break;
 						}
 
@@ -3183,18 +3624,24 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 
 						if (!evaluateExpr(thread, st, expr, &result))
 						{
-							commandFailed(st, argv[0], "evaluation of meta-command failed");
-							st->state = CSTATE_ABORTED;
+							commandFailed(st, argv[0],
+										  "evaluation of meta-command failed",
+										  false);
+							st->failure_status = META_COMMAND_FAILURE;
+							st->state = CSTATE_FAILURE;
 							break;
 						}
 
 						if (command->meta == META_SET)
 						{
-							if (!putVariableValue(&st->variables, argv[0],
-												  argv[1], &result,  true))
+							if (!putVariableValue(&st->variables,  argv[0],
+												  argv[1], &result, true))
 							{
-								commandFailed(st, "set", "assignment of meta-command failed");
-								st->state = CSTATE_ABORTED;
+								commandFailed(st, "set",
+											  "assignment of meta-command failed",
+											  false);
+								st->failure_status = META_COMMAND_FAILURE;
+								st->state = CSTATE_FAILURE;
 								break;
 							}
 						}
@@ -3255,8 +3702,11 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 						}
 						else if (!ret)	/* on error */
 						{
-							commandFailed(st, "setshell", "execution of meta-command failed");
-							st->state = CSTATE_ABORTED;
+							commandFailed(st, "setshell",
+										  "execution of meta-command failed",
+										  false);
+							st->failure_status = META_COMMAND_FAILURE;
+							st->state = CSTATE_FAILURE;
 							break;
 						}
 						else
@@ -3276,8 +3726,11 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 						}
 						else if (!ret)	/* on error */
 						{
-							commandFailed(st, "shell", "execution of meta-command failed");
-							st->state = CSTATE_ABORTED;
+							commandFailed(st, "shell",
+										  "execution of meta-command failed",
+										  false);
+							st->failure_status = META_COMMAND_FAILURE;
+							st->state = CSTATE_FAILURE;
 							break;
 						}
 						else
@@ -3393,36 +3846,54 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 				 * Wait for the current SQL command to complete
 				 */
 			case CSTATE_WAIT_RESULT:
-				command = sql_script[st->use_file].commands[st->command];
-				pgbench_error(DEBUG, "client %d receiving\n", st->id);
-				if (!PQconsumeInput(st->con))
-				{				/* there's something wrong */
-					commandFailed(st, "SQL", "perhaps the backend died while processing");
-					st->state = CSTATE_ABORTED;
-					break;
-				}
-				if (PQisBusy(st->con))
-					return;		/* don't have the whole result yet */
-
-				/*
-				 * Read and discard the query result;
-				 */
-				res = PQgetResult(st->con);
-				switch (PQresultStatus(res))
 				{
-					case PGRES_COMMAND_OK:
-					case PGRES_TUPLES_OK:
-					case PGRES_EMPTY_QUERY:
-						/* OK */
-						PQclear(res);
-						discard_response(st);
-						st->state = CSTATE_END_COMMAND;
-						break;
-					default:
-						commandFailed(st, "SQL", PQerrorMessage(st->con));
-						PQclear(res);
+					char	   *sqlState;
+
+					command = sql_script[st->use_file].commands[st->command];
+					pgbench_error(DEBUG, "client %d receiving\n", st->id);
+					if (!PQconsumeInput(st->con))
+					{				/* there's something wrong */
+						commandFailed(st, "SQL",
+									  "perhaps the backend died while processing",
+									  true);
 						st->state = CSTATE_ABORTED;
 						break;
+					}
+					if (PQisBusy(st->con))
+						return;		/* don't have the whole result yet */
+
+					/*
+					 * Read and discard the query result;
+					 */
+					res = PQgetResult(st->con);
+					sqlState = PQresultErrorField(res, PG_DIAG_SQLSTATE);
+					switch (PQresultStatus(res))
+					{
+						case PGRES_COMMAND_OK:
+						case PGRES_TUPLES_OK:
+						case PGRES_EMPTY_QUERY:
+							/* OK */
+							st->failure_status = NO_FAILURE;
+							PQclear(res);
+							discard_response(st);
+							st->state = CSTATE_END_COMMAND;
+							break;
+						case PGRES_NONFATAL_ERROR:
+						case PGRES_FATAL_ERROR:
+							st->failure_status = getSQLFailureStatus(sqlState);
+							commandFailed(st, "SQL", PQerrorMessage(st->con),
+										  false);
+							PQclear(res);
+							discard_response(st);
+							st->state = CSTATE_FAILURE;
+							break;
+						default:
+							commandFailed(st, "SQL", PQerrorMessage(st->con),
+										  true);
+							PQclear(res);
+							st->state = CSTATE_ABORTED;
+							break;
+					}
 				}
 				break;
 
@@ -3451,7 +3922,7 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 				 * in thread-local data structure, if per-command latencies
 				 * are requested.
 				 */
-				if (is_latencies)
+				if (report_per_command)
 				{
 					if (INSTR_TIME_IS_ZERO(now))
 						INSTR_TIME_SET_CURRENT(now);
@@ -3469,6 +3940,169 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 					CSTATE_START_COMMAND : CSTATE_SKIP_COMMAND;
 				break;
 
+			/*
+			 * Clean up after a failed transaction.
+			 */
+			case CSTATE_FAILURE:
+				{
+					bool		in_tx_block;
+
+					Assert(st->failure_status != NO_FAILURE);
+
+					/*
+					 * Check if we have a failed transaction block or not, and
+					 * roll it back if any.
+					 */
+					in_tx_block = inTransactionBlock(st);
+					if (st->state == CSTATE_ABORTED)
+						break;		/* there's something wrong */
+					if (in_tx_block)
+					{
+						PGresult   *res;
+
+						res = PQexec(st->con, "ROLLBACK");
+						if (PQresultStatus(res) != PGRES_COMMAND_OK)
+						{
+							/*
+							 * We are sure that the function PQerrorMessage is
+							 * always called.
+							 */
+							Assert(LOG >= log_min_messages);
+
+							pgbench_error(LOG,
+										  "client %d aborted during the termination of the failed transaction block; %s\n",
+										  st->id, PQerrorMessage(st->con));
+							PQclear(res);
+							st->state = CSTATE_ABORTED;
+							break;
+						}
+						PQclear(res);
+					}
+
+					/* Clear the conditional stack */
+					conditional_stack_reset(st->cstack);
+
+					/* Check if we can retry the failure */
+					st->state = CSTATE_RETRY;
+				}
+				break;
+
+			/*
+			 * Retry the failed transaction if possible.
+			 */
+			case CSTATE_RETRY:
+				command = sql_script[st->use_file].commands[st->command];
+
+				if (canRetry(st, &now))
+				{
+					/*
+					 * The failed transaction will be retried. So accumulate
+					 * the retry.
+					 */
+					st->retries++;
+					command->retries++;
+
+					/*
+					 * Inform that the failed transaction will be retried.
+					 * Allocate memory for the message only if necessary.
+					 */
+					if (DEBUG_FAIL >= log_min_messages)
+					{
+						PQExpBufferData errmsg_buf;
+
+						initPQExpBuffer(&errmsg_buf);
+						printfPQExpBuffer(&errmsg_buf,
+										  "client %d repeats the failed transaction (try %d",
+										  st->id, st->retries + 1);
+						if (max_tries)
+							appendPQExpBuffer(&errmsg_buf, "/%d", max_tries);
+						if (latency_limit)
+						{
+							appendPQExpBuffer(&errmsg_buf,
+											  ", %.3f%% of the maximum time of tries was used",
+											  getLatencyUsed(st, &now));
+						}
+						appendPQExpBufferStr(&errmsg_buf, ")\n");
+						pgbench_error(DEBUG_FAIL, "%s", errmsg_buf.data);
+						termPQExpBuffer(&errmsg_buf);
+					}
+
+					/*
+					 * Reset the execution parameters as they were at the
+					 * beginning of the transaction.
+					 */
+					memcpy(&(st->random_state), &(st->retry_state.random_state),
+						   sizeof(RandomState));
+					if (!copyVariables(&st->variables,
+									   &st->retry_state.variables))
+					{
+						pgbench_error(LOG,
+									  "client %d aborted when preparing to retry the failed transaction\n",
+									  st->id);
+						st->state = CSTATE_ABORTED;
+						break;
+					}
+
+					/* Process the first transaction command */
+					st->command = 0;
+					st->failure_status = NO_FAILURE;
+					st->state = CSTATE_START_COMMAND;
+				}
+				else
+				{
+					/*
+					 * We will not be able to retry this failed transaction.
+					 * So accumulate the error.
+					 */
+					command->ecnt++;
+
+					/*
+					 * If this is a serialization or deadlock failure, inform
+					 * that the failed transaction will not be retried. Allocate
+					 * memory for the message only if necessary.
+					 */
+					if (DEBUG_FAIL >= log_min_messages &&
+						canRetryFailure(st->failure_status))
+					{
+						PQExpBufferData errmsg_buf;
+
+						initPQExpBuffer(&errmsg_buf);
+						printfPQExpBuffer(&errmsg_buf,
+										  "client %d ends the failed transaction (try %d",
+										  st->id, st->retries + 1);
+						if (max_tries)
+							appendPQExpBuffer(&errmsg_buf, "/%d", max_tries);
+						if (latency_limit)
+						{
+							appendPQExpBuffer(&errmsg_buf,
+											  ", %.3f%% of the maximum time of tries was used",
+											  getLatencyUsed(st, &now));
+						}
+						appendPQExpBufferStr(&errmsg_buf, ")\n");
+						pgbench_error(DEBUG_FAIL, "%s", errmsg_buf.data);
+						termPQExpBuffer(&errmsg_buf);
+					}
+
+					/*
+					 * Reset the execution parameters as they were at the
+					 * beginning of the transaction except for a random
+					 * state.
+					 */
+					if (!copyVariables(&st->variables,
+									   &st->retry_state.variables))
+					{
+						pgbench_error(LOG,
+									  "client %d aborted when ending the failed transaction\n",
+									  st->id);
+						st->state = CSTATE_ABORTED;
+						break;
+					}
+
+					/* End the failed transaction */
+					st->state = CSTATE_END_TX;
+				}
+				break;
+
 				/*
 				 * End of transaction.
 				 */
@@ -3490,7 +4124,8 @@ doCustom(TState *thread, CState *st, StatsData *agg)
 					INSTR_TIME_SET_ZERO(now);
 				}
 
-				if ((st->cnt >= nxacts && duration <= 0) || timer_exceeded)
+				if ((getTotalCnt(st) >= nxacts && duration <= 0) ||
+					timer_exceeded)
 				{
 					/* exit success */
 					st->state = CSTATE_FINISHED;
@@ -3573,6 +4208,20 @@ doLog(TState *thread, CState *st,
 					agg->latency.sum2,
 					agg->latency.min,
 					agg->latency.max);
+
+			if (errors_detailed)
+			{
+				fprintf(logfile, " " INT64_FORMAT " " INT64_FORMAT " " INT64_FORMAT " " INT64_FORMAT,
+						agg->serialization_errors,
+						agg->deadlock_errors,
+						agg->other_sql_errors,
+						agg->meta_command_errors);
+			}
+			else
+			{
+				fprintf(logfile, " " INT64_FORMAT, getEcnt(agg));
+			}
+
 			if (throttle_delay)
 			{
 				fprintf(logfile, " %.0f %.0f %.0f %.0f",
@@ -3583,6 +4232,10 @@ doLog(TState *thread, CState *st,
 				if (latency_limit)
 					fprintf(logfile, " " INT64_FORMAT, agg->skipped);
 			}
+			if (RETRIES_ENABLED)
+				fprintf(logfile, " " INT64_FORMAT " " INT64_FORMAT,
+						agg->retried,
+						agg->retries);
 			fputc('\n', logfile);
 
 			/* reset data and move to next interval */
@@ -3590,7 +4243,7 @@ doLog(TState *thread, CState *st,
 		}
 
 		/* accumulate the current transaction */
-		accumStats(agg, skipped, latency, lag);
+		accumStats(agg, skipped, latency, lag, st->failure_status, st->retries);
 	}
 	else
 	{
@@ -3600,14 +4253,46 @@ doLog(TState *thread, CState *st,
 		gettimeofday(&tv, NULL);
 		if (skipped)
 			fprintf(logfile, "%d " INT64_FORMAT " skipped %d %ld %ld",
-					st->id, st->cnt, st->use_file,
+					st->id, getTotalCnt(st), st->use_file,
 					(long) tv.tv_sec, (long) tv.tv_usec);
-		else
+		else if (st->failure_status == NO_FAILURE)
 			fprintf(logfile, "%d " INT64_FORMAT " %.0f %d %ld %ld",
-					st->id, st->cnt, latency, st->use_file,
+					st->id, getTotalCnt(st), latency, st->use_file,
+					(long) tv.tv_sec, (long) tv.tv_usec);
+		else if (errors_detailed)
+		{
+			if (st->failure_status == SERIALIZATION_FAILURE)
+				fprintf(logfile, "%d " INT64_FORMAT " serialization_error %d %ld %ld",
+						st->id, getTotalCnt(st), st->use_file,
+						(long) tv.tv_sec, (long) tv.tv_usec);
+			else if (st->failure_status == DEADLOCK_FAILURE)
+				fprintf(logfile, "%d " INT64_FORMAT " deadlock_error %d %ld %ld",
+						st->id, getTotalCnt(st), st->use_file,
+						(long) tv.tv_sec, (long) tv.tv_usec);
+			else if (st->failure_status == ANOTHER_SQL_FAILURE)
+				fprintf(logfile, "%d " INT64_FORMAT " another_sql_error %d %ld %ld",
+						st->id, getTotalCnt(st), st->use_file,
+						(long) tv.tv_sec, (long) tv.tv_usec);
+			else if (st->failure_status == META_COMMAND_FAILURE)
+				fprintf(logfile, "%d " INT64_FORMAT " meta_command_error %d %ld %ld",
+						st->id, getTotalCnt(st), st->use_file,
+						(long) tv.tv_sec, (long) tv.tv_usec);
+			else
+			{
+				/* internal error which should never occur */
+				pgbench_error(ERROR, "unexpected failure status: %d",
+							  st->failure_status);
+			}
+		}
+		else
+			fprintf(logfile, "%d " INT64_FORMAT " failed %d %ld %ld",
+					st->id, getTotalCnt(st), st->use_file,
 					(long) tv.tv_sec, (long) tv.tv_usec);
+
 		if (throttle_delay)
 			fprintf(logfile, " %.0f", lag);
+		if (RETRIES_ENABLED)
+			fprintf(logfile, " %d", st->retries);
 		fputc('\n', logfile);
 	}
 }
@@ -3624,10 +4309,11 @@ processXactStats(TState *thread, CState *st, instr_time *now,
 {
 	double		latency = 0.0,
 				lag = 0.0;
-	bool		thread_details = progress || throttle_delay || latency_limit,
-				detailed = thread_details || use_log || per_script_stats;
+	bool		detailed = progress || throttle_delay || latency_limit ||
+						   use_log || per_script_stats;
 
-	if (detailed && !skipped)
+	if (detailed && !skipped &&
+		(st->failure_status == NO_FAILURE || latency_limit))
 	{
 		if (INSTR_TIME_IS_ZERO(*now))
 			INSTR_TIME_SET_CURRENT(*now);
@@ -3637,30 +4323,29 @@ processXactStats(TState *thread, CState *st, instr_time *now,
 		lag = INSTR_TIME_GET_MICROSEC(st->txn_begin) - st->txn_scheduled;
 	}
 
-	if (thread_details)
-	{
-		/* keep detailed thread stats */
-		accumStats(&thread->stats, skipped, latency, lag);
+	/* keep detailed thread stats */
+	accumStats(&thread->stats, skipped, latency, lag, st->failure_status,
+			   st->retries);
 
-		/* count transactions over the latency limit, if needed */
-		if (latency_limit && latency > latency_limit)
-			thread->latency_late++;
-	}
-	else
-	{
-		/* no detailed stats, just count */
-		thread->stats.cnt++;
-	}
+	/* count transactions over the latency limit, if needed */
+	if (latency_limit && latency > latency_limit)
+		thread->latency_late++;
 
 	/* client stat is just counting */
-	st->cnt++;
+	if (st->failure_status == NO_FAILURE)
+		st->cnt++;
+	else
+		st->ecnt++;
 
 	if (use_log)
 		doLog(thread, st, agg, skipped, latency, lag);
 
 	/* XXX could use a mutex here, but we choose not to */
 	if (per_script_stats)
-		accumStats(&sql_script[st->use_file].stats, skipped, latency, lag);
+	{
+		accumStats(&sql_script[st->use_file].stats, skipped, latency, lag,
+				   st->failure_status, st->retries);
+	}
 }
 
 
@@ -4821,7 +5506,9 @@ printResults(TState *threads, StatsData *total, instr_time total_time,
 	double		time_include,
 				tps_include,
 				tps_exclude;
-	int64		ntx = total->cnt - total->skipped;
+	int64		ecnt = getEcnt(total);
+	int64		ntx = total->cnt - total->skipped,
+				total_ntx = total->cnt + ecnt;
 	int			i,
 				totalCacheOverflows = 0;
 
@@ -4842,8 +5529,8 @@ printResults(TState *threads, StatsData *total, instr_time total_time,
 	if (duration <= 0)
 	{
 		printf("number of transactions per client: %d\n", nxacts);
-		printf("number of transactions actually processed: " INT64_FORMAT "/%d\n",
-			   ntx, nxacts * nclients);
+		printf("number of transactions actually processed: " INT64_FORMAT "/" INT64_FORMAT "\n",
+			   ntx, total_ntx);
 	}
 	else
 	{
@@ -4851,6 +5538,67 @@ printResults(TState *threads, StatsData *total, instr_time total_time,
 		printf("number of transactions actually processed: " INT64_FORMAT "\n",
 			   ntx);
 	}
+
+	if (ecnt > 0)
+	{
+		printf("number of errors: " INT64_FORMAT " (%.3f%%)\n",
+			   ecnt, 100.0 * ecnt / total_ntx);
+
+		if (errors_detailed)
+		{
+			/* SQL errors */
+			if (total->serialization_errors || total->other_sql_errors)
+			{
+				printf("number of serialization errors: " INT64_FORMAT " (%.3f%%)\n",
+					   total->serialization_errors,
+					   100.0 * total->serialization_errors / total_ntx);
+			}
+			if (total->deadlock_errors || total->other_sql_errors)
+			{
+				printf("number of deadlock errors: " INT64_FORMAT " (%.3f%%)\n",
+					   total->deadlock_errors,
+					   100.0 * total->deadlock_errors / total_ntx);
+			}
+			if (total->other_sql_errors)
+			{
+				printf("number of other SQL errors: " INT64_FORMAT " (%.3f%%)\n",
+					   total->other_sql_errors,
+					   100.0 * total->other_sql_errors / total_ntx);
+			}
+
+			/* errors in meta commands */
+			if (total->meta_command_errors > 0)
+			{
+				printf("number of errors in meta-commands: " INT64_FORMAT " (%.3f%%)\n",
+					   total->meta_command_errors,
+					   100.0 * total->meta_command_errors / total_ntx);
+			}
+		}
+	}
+
+	/*
+	 * It can be non-zero only if max_tries is greater than one or
+	 * latency_limit is used.
+	 */
+	if (total->retried > 0)
+	{
+		printf("number of retried: " INT64_FORMAT " (%.3f%%)\n",
+			   total->retried, 100.0 * total->retried / total_ntx);
+		printf("number of retries: " INT64_FORMAT "\n", total->retries);
+	}
+
+	if (max_tries)
+		printf("maximum number of tries: %d\n", max_tries);
+
+	if (latency_limit)
+	{
+		/* this statistics includes both successful and failed transactions */
+		printf("number of transactions above the %.1f ms latency limit: " INT64_FORMAT "/" INT64_FORMAT " (%.3f %%)%s\n",
+			   latency_limit / 1000.0, latency_late, total_ntx,
+			   (total_ntx > 0) ? 100.0 * latency_late / total_ntx : 0.0,
+			   ecnt > 0 ? " (including errors)" : "");
+	}
+
 	/* Report zipfian cache overflow */
 	for (i = 0; i < nthreads; i++)
 	{
@@ -4870,18 +5618,14 @@ printResults(TState *threads, StatsData *total, instr_time total_time,
 			   total->skipped,
 			   100.0 * total->skipped / total->cnt);
 
-	if (latency_limit)
-		printf("number of transactions above the %.1f ms latency limit: " INT64_FORMAT "/" INT64_FORMAT " (%.3f %%)\n",
-			   latency_limit / 1000.0, latency_late, ntx,
-			   (ntx > 0) ? 100.0 * latency_late / ntx : 0.0);
-
 	if (throttle_delay || progress || latency_limit)
 		printSimpleStats("latency", &total->latency);
 	else
 	{
 		/* no measurement, show average latency computed from run time */
-		printf("latency average = %.3f ms\n",
-			   1000.0 * time_include * nclients / total->cnt);
+		printf("latency average = %.3f ms%s\n",
+			   1000.0 * time_include * nclients / total_ntx,
+			   ecnt > 0 ? " (including errors)" : "");
 	}
 
 	if (throttle_delay)
@@ -4900,7 +5644,7 @@ printResults(TState *threads, StatsData *total, instr_time total_time,
 	printf("tps = %f (excluding connections establishing)\n", tps_exclude);
 
 	/* Report per-script/command statistics */
-	if (per_script_stats || is_latencies)
+	if (per_script_stats || report_per_command)
 	{
 		int			i;
 
@@ -4909,6 +5653,8 @@ printResults(TState *threads, StatsData *total, instr_time total_time,
 			if (per_script_stats)
 			{
 				StatsData  *sstats = &sql_script[i].stats;
+				int64		script_ecnt = getEcnt(sstats);
+				int64		script_total_ntx = sstats->cnt + script_ecnt;
 
 				printf("SQL script %d: %s\n"
 					   " - weight: %d (targets %.1f%% of total)\n"
@@ -4920,6 +5666,62 @@ printResults(TState *threads, StatsData *total, instr_time total_time,
 					   100.0 * sstats->cnt / total->cnt,
 					   (sstats->cnt - sstats->skipped) / time_include);
 
+				if (ecnt > 0)
+				{
+					printf(" - number of errors: " INT64_FORMAT " (%.3f%%)\n",
+						   script_ecnt, 100.0 * script_ecnt / script_total_ntx);
+
+					if (errors_detailed)
+					{
+						/* SQL errors */
+						if (total->serialization_errors ||
+							total->other_sql_errors)
+						{
+							printf(" - number of serialization errors: " INT64_FORMAT " (%.3f%%)\n",
+								   sstats->serialization_errors,
+								   (100.0 * sstats->serialization_errors /
+									script_total_ntx));
+						}
+						if (total->deadlock_errors ||
+							total->other_sql_errors)
+						{
+							printf(" - number of deadlock errors: " INT64_FORMAT " (%.3f%%)\n",
+								   sstats->deadlock_errors,
+								   (100.0 * sstats->deadlock_errors /
+									script_total_ntx));
+						}
+						if (total->other_sql_errors)
+						{
+							printf(" - number of other SQL errors: " INT64_FORMAT " (%.3f%%)\n",
+								   sstats->other_sql_errors,
+								   (100.0 * sstats->other_sql_errors /
+									script_total_ntx));
+						}
+
+						/* errors in meta commands */
+						if (total->meta_command_errors > 0)
+						{
+							printf(" - number of errors in meta-commands: " INT64_FORMAT " (%.3f%%)\n",
+								   sstats->meta_command_errors,
+								   (100.0 * sstats->meta_command_errors /
+									script_total_ntx));
+						}
+					}
+				}
+
+				/*
+				 * It can be non-zero only if max_tries is greater than one or
+				 * latency_limit is used.
+				 */
+				if (total->retried > 0)
+				{
+					printf(" - number of retried: " INT64_FORMAT " (%.3f%%)\n",
+						   sstats->retried,
+						   100.0 * sstats->retried / script_total_ntx);
+					printf(" - number of retries: " INT64_FORMAT "\n",
+						   sstats->retries);
+				}
+
 				if (throttle_delay && latency_limit && sstats->cnt > 0)
 					printf(" - number of transactions skipped: " INT64_FORMAT " (%.3f%%)\n",
 						   sstats->skipped,
@@ -4928,15 +5730,16 @@ printResults(TState *threads, StatsData *total, instr_time total_time,
 				printSimpleStats(" - latency", &sstats->latency);
 			}
 
-			/* Report per-command latencies */
-			if (is_latencies)
+			/* Report per-command latencies and errors */
+			if (report_per_command)
 			{
 				Command   **commands;
 
-				if (per_script_stats)
-					printf(" - statement latencies in milliseconds:\n");
-				else
-					printf("statement latencies in milliseconds:\n");
+				printf("%sstatement latencies in milliseconds%s:\n",
+					   per_script_stats ? " - " : "",
+					   (RETRIES_ENABLED ?
+						", errors and retries" :
+						" and errors"));
 
 				for (commands = sql_script[i].commands;
 					 *commands != NULL;
@@ -4944,10 +5747,23 @@ printResults(TState *threads, StatsData *total, instr_time total_time,
 				{
 					SimpleStats *cstats = &(*commands)->stats;
 
-					printf("   %11.3f  %s\n",
-						   (cstats->count > 0) ?
-						   1000.0 * cstats->sum / cstats->count : 0.0,
-						   (*commands)->line);
+					if (RETRIES_ENABLED)
+					{
+						printf("   %11.3f  %10" INT64_MODIFIER "d  %10" INT64_MODIFIER "d  %s\n",
+							   (cstats->count > 0) ?
+							   1000.0 * cstats->sum / cstats->count : 0.0,
+							   (*commands)->ecnt,
+							   (*commands)->retries,
+							   (*commands)->line);
+					}
+					else
+					{
+						printf("   %11.3f  %10" INT64_MODIFIER "d  %s\n",
+							   (cstats->count > 0) ?
+							   1000.0 * cstats->sum / cstats->count : 0.0,
+							   (*commands)->ecnt,
+							   (*commands)->line);
+					}
 				}
 			}
 		}
@@ -5038,7 +5854,7 @@ main(int argc, char **argv)
 		{"progress", required_argument, NULL, 'P'},
 		{"protocol", required_argument, NULL, 'M'},
 		{"quiet", no_argument, NULL, 'q'},
-		{"report-latencies", no_argument, NULL, 'r'},
+		{"report-per-command", no_argument, NULL, 'r'},
 		{"rate", required_argument, NULL, 'R'},
 		{"scale", required_argument, NULL, 's'},
 		{"select-only", no_argument, NULL, 'S'},
@@ -5057,6 +5873,9 @@ main(int argc, char **argv)
 		{"log-prefix", required_argument, NULL, 7},
 		{"foreign-keys", no_argument, NULL, 8},
 		{"random-seed", required_argument, NULL, 9},
+		{"debug-fails", no_argument, NULL, 10},
+		{"errors-detailed", no_argument, NULL, 11},
+		{"max-tries", required_argument, NULL, 12},
 		{NULL, 0, NULL, 0}
 	};
 
@@ -5218,7 +6037,7 @@ main(int argc, char **argv)
 				break;
 			case 'r':
 				benchmarking_option_set = true;
-				is_latencies = true;
+				report_per_command = true;
 				break;
 			case 's':
 				scale_given = true;
@@ -5409,6 +6228,28 @@ main(int argc, char **argv)
 								  "error while setting random seed from --random-seed option\n");
 				}
 				break;
+			case 10:			/* debug-fails */
+				/* do not conflict with the option --debug */
+				if (log_min_messages > DEBUG_FAIL)
+					log_min_messages = DEBUG_FAIL;
+				break;
+			case 11:			/* errors-detailed */
+				errors_detailed = true;
+				break;
+			case 12:			/* max-tries */
+				{
+					int32		max_tries_arg = atoi(optarg);
+
+					if (max_tries_arg <= 0)
+					{
+						pgbench_error(ERROR,
+									  "invalid number of maximum tries: \"%s\"\n",
+									  optarg);
+					}
+					benchmarking_option_set = true;
+					max_tries = (uint32) max_tries_arg;
+				}
+				break;
 			default:
 				pgbench_error(ERROR,
 							  _("Try \"%s --help\" for more information.\n"),
@@ -5578,6 +6419,10 @@ main(int argc, char **argv)
 					  "--progress-timestamp is allowed only under --progress\n");
 	}
 
+	/* If necessary set the default tries limit  */
+	if (!max_tries && !latency_limit)
+		max_tries = 1;
+
 	/*
 	 * save main process id in the global variable because process id will be
 	 * changed after fork.
@@ -5861,6 +6706,12 @@ main(int argc, char **argv)
 		mergeSimpleStats(&stats.lag, &thread->stats.lag);
 		stats.cnt += thread->stats.cnt;
 		stats.skipped += thread->stats.skipped;
+		stats.retries += thread->stats.retries;
+		stats.retried += thread->stats.retried;
+		stats.serialization_errors += thread->stats.serialization_errors;
+		stats.deadlock_errors += thread->stats.deadlock_errors;
+		stats.other_sql_errors += thread->stats.other_sql_errors;
+		stats.meta_command_errors += thread->stats.meta_command_errors;
 		latency_late += thread->latency_late;
 		INSTR_TIME_ADD(conn_total_time, thread->conn_time);
 	}
@@ -6164,7 +7015,10 @@ threadRun(void *arg)
 				/* generate and show report */
 				StatsData	cur;
 				int64		run = now - last_report,
-							ntx;
+							ntx,
+							retries,
+							retried,
+							ecnt;
 				double		tps,
 							total_run,
 							latency,
@@ -6192,6 +7046,14 @@ threadRun(void *arg)
 					mergeSimpleStats(&cur.lag, &thread[i].stats.lag);
 					cur.cnt += thread[i].stats.cnt;
 					cur.skipped += thread[i].stats.skipped;
+					cur.retries += thread[i].stats.retries;
+					cur.retried += thread[i].stats.retried;
+					cur.serialization_errors +=
+						thread[i].stats.serialization_errors;
+					cur.deadlock_errors += thread[i].stats.deadlock_errors;
+					cur.other_sql_errors += thread[i].stats.other_sql_errors;
+					cur.meta_command_errors +=
+						thread[i].stats.meta_command_errors;
 				}
 
 				/* we count only actually executed transactions */
@@ -6209,6 +7071,9 @@ threadRun(void *arg)
 				{
 					latency = sqlat = stdev = lag = 0;
 				}
+				retries = cur.retries - last.retries;
+				retried = cur.retried - last.retried;
+				ecnt = getEcnt(&cur) - getEcnt(&last);
 
 				if (progress_timestamp)
 				{
@@ -6241,6 +7106,12 @@ threadRun(void *arg)
 								  "progress: %s, %.1f tps, lat %.3f ms stddev %.3f",
 								  tbuf, tps, latency, stdev);
 
+				if (ecnt > 0)
+				{
+					appendPQExpBuffer(&progress_buf,
+									  ", " INT64_FORMAT " failed", ecnt);
+				}
+
 				if (throttle_delay)
 				{
 					appendPQExpBuffer(&progress_buf, ", lag %.3f ms", lag);
@@ -6249,6 +7120,17 @@ threadRun(void *arg)
 										  ", " INT64_FORMAT " skipped",
 										  cur.skipped - last.skipped);
 				}
+
+				/*
+				 * It can be non-zero only if max_tries is greater than one or
+				 * latency_limit is used.
+				 */
+				if (retried > 0)
+				{
+					appendPQExpBuffer(&progress_buf,
+									  ", " INT64_FORMAT " retried, " INT64_FORMAT " retries",
+									  retried, retries);
+				}
 				appendPQExpBufferChar(&progress_buf, '\n');
 
 				pgbench_error(LOG, "%s", progress_buf.data);
diff --git a/src/bin/pgbench/t/001_pgbench_with_server.pl b/src/bin/pgbench/t/001_pgbench_with_server.pl
index 2fc021d..2114329 100644
--- a/src/bin/pgbench/t/001_pgbench_with_server.pl
+++ b/src/bin/pgbench/t/001_pgbench_with_server.pl
@@ -5,9 +5,20 @@ use PostgresNode;
 use TestLib;
 use Test::More;
 
+use constant
+{
+	SQL_ERROR           => 0,
+	META_COMMAND_ERROR  => 1,
+	SYNTAX_ERROR        => 2,
+};
+
 # start a pgbench specific server
 my $node = get_new_node('main');
-$node->init;
+
+# Set to untranslated messages, to be able to compare program output with
+# expected strings.
+$node->init(extra => [ '--locale', 'C' ]);
+
 $node->start;
 
 # invoke pgbench
@@ -136,7 +147,8 @@ pgbench(
 		qr{builtin: TPC-B},
 		qr{clients: 2\b},
 		qr{processed: 10/10},
-		qr{mode: simple}
+		qr{mode: simple},
+		qr{maximum number of tries: 1}
 	],
 	[qr{^$}],
 	'pgbench tpcb-like');
@@ -530,11 +542,12 @@ pgbench(
 # trigger many expression errors
 my @errors = (
 
-	# [ test name, expected status, expected stderr, script ]
+	# [ test name, expected status, error type, expected stderr, script ]
 	# SQL
 	[
 		'sql syntax error',
 		0,
+		SQL_ERROR,
 		[
 			qr{ERROR:  syntax error},
 			qr{prepared statement .* does not exist}
@@ -544,28 +557,36 @@ my @errors = (
 }
 	],
 	[
-		'sql too many args', 1, [qr{statement has too many arguments.*\b9\b}],
+		'sql too many args', 1, SYNTAX_ERROR,
+		[qr{statement has too many arguments.*\b9\b}],
 		q{-- MAX_ARGS=10 for prepared
 \set i 0
 SELECT LEAST(:i, :i, :i, :i, :i, :i, :i, :i, :i, :i, :i);
 }
 	],
+	[   'sql division by zero', 0, SQL_ERROR, [qr{ERROR:  division by zero}],
+		q{-- SQL division by zero
+SELECT 1 / 0;
+}
+	],
 
 	# SHELL
 	[
-		'shell bad command',                    0,
+		'shell bad command', 0, META_COMMAND_ERROR,
 		[qr{\(shell\) .* meta-command failed}], q{\shell no-such-command}
 	],
 	[
-		'shell undefined variable', 0,
+		'shell undefined variable', 0, META_COMMAND_ERROR,
 		[qr{undefined variable ":nosuchvariable"}],
 		q{-- undefined variable in shell
 \shell echo ::foo :nosuchvariable
 }
 	],
-	[ 'shell missing command', 1, [qr{missing command }], q{\shell} ],
+	[   'shell missing command', 1, SYNTAX_ERROR, [qr{missing command }],
+		q{\shell} ],
 	[
-		'shell too many args', 1, [qr{too many arguments in command "shell"}],
+		'shell too many args', 1, SYNTAX_ERROR,
+		[qr{too many arguments in command "shell"}],
 		q{-- 257 arguments to \shell
 \shell echo \
  0 1 2 3 4 5 6 7 8 9 A B C D E F \
@@ -589,162 +610,232 @@ SELECT LEAST(:i, :i, :i, :i, :i, :i, :i, :i, :i, :i, :i);
 
 	# SET
 	[
-		'set syntax error',                  1,
+		'set syntax error', 1, SYNTAX_ERROR,
 		[qr{syntax error in command "set"}], q{\set i 1 +}
 	],
 	[
-		'set no such function',         1,
+		'set no such function', 1, SYNTAX_ERROR,
 		[qr{unexpected function name}], q{\set i noSuchFunction()}
 	],
 	[
-		'set invalid variable name', 0,
+		'set invalid variable name', 0, META_COMMAND_ERROR,
 		[qr{invalid variable name}], q{\set . 1}
 	],
 	[
-		'set int overflow',                   0,
+		'set int overflow', 0, META_COMMAND_ERROR,
 		[qr{double to int overflow for 100}], q{\set i int(1E32)}
 	],
-	[ 'set division by zero', 0, [qr{division by zero}], q{\set i 1/0} ],
 	[
-		'set bigint out of range', 0,
+		'set division by zero', 0, META_COMMAND_ERROR,
+		[qr{division by zero}], q{\set i 1/0}
+	],
+	[
+		'set bigint out of range', 0, META_COMMAND_ERROR,
 		[qr{bigint out of range}], q{\set i 9223372036854775808 / -1}
 	],
 	[
 		'set undefined variable',
 		0,
+		META_COMMAND_ERROR,
 		[qr{undefined variable "nosuchvariable"}],
 		q{\set i :nosuchvariable}
 	],
-	[ 'set unexpected char', 1, [qr{unexpected character .;.}], q{\set i ;} ],
+	[
+		'set unexpected char', 1, SYNTAX_ERROR,
+		[qr{unexpected character .;.}], q{\set i ;}
+	],
 	[
 		'set too many args',
 		0,
+		META_COMMAND_ERROR,
 		[qr{too many function arguments}],
 		q{\set i least(0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)}
 	],
 	[
-		'set empty random range',          0,
+		'set empty random range', 0, META_COMMAND_ERROR,
 		[qr{empty range given to random}], q{\set i random(5,3)}
 	],
 	[
 		'set random range too large',
 		0,
+		META_COMMAND_ERROR,
 		[qr{random range is too large}],
 		q{\set i random(-9223372036854775808, 9223372036854775807)}
 	],
 	[
 		'set gaussian param too small',
 		0,
+		META_COMMAND_ERROR,
 		[qr{gaussian param.* at least 2}],
 		q{\set i random_gaussian(0, 10, 1.0)}
 	],
 	[
 		'set exponential param greater 0',
 		0,
+		META_COMMAND_ERROR,
 		[qr{exponential parameter must be greater }],
 		q{\set i random_exponential(0, 10, 0.0)}
 	],
 	[
 		'set zipfian param to 1',
 		0,
+		META_COMMAND_ERROR,
 		[qr{zipfian parameter must be in range \(0, 1\) U \(1, \d+\]}],
 		q{\set i random_zipfian(0, 10, 1)}
 	],
 	[
 		'set zipfian param too large',
 		0,
+		META_COMMAND_ERROR,
 		[qr{zipfian parameter must be in range \(0, 1\) U \(1, \d+\]}],
 		q{\set i random_zipfian(0, 10, 1000000)}
 	],
 	[
-		'set non numeric value',                     0,
+		'set non numeric value', 0, META_COMMAND_ERROR,
 		[qr{malformed variable "foo" value: "bla"}], q{\set i :foo + 1}
 	],
-	[ 'set no expression',    1, [qr{syntax error}],      q{\set i} ],
-	[ 'set missing argument', 1, [qr{missing argument}i], q{\set} ],
+	[ 'set no expression', 1, SYNTAX_ERROR, [qr{syntax error}], q{\set i} ],
 	[
-		'set not a bool',                      0,
+		'set missing argument', 1, SYNTAX_ERROR,
+		[qr{missing argument}i], q{\set}
+	],
+	[
+		'set not a bool', 0, META_COMMAND_ERROR,
 		[qr{cannot coerce double to boolean}], q{\set b NOT 0.0}
 	],
 	[
-		'set not an int',                   0,
+		'set not an int', 0, META_COMMAND_ERROR,
 		[qr{cannot coerce boolean to int}], q{\set i TRUE + 2}
 	],
 	[
-		'set not a double',                    0,
+		'set not a double', 0, META_COMMAND_ERROR,
 		[qr{cannot coerce boolean to double}], q{\set d ln(TRUE)}
 	],
 	[
 		'set case error',
 		1,
+		SYNTAX_ERROR,
 		[qr{syntax error in command "set"}],
 		q{\set i CASE TRUE THEN 1 ELSE 0 END}
 	],
 	[
-		'set random error',                 0,
+		'set random error', 0, META_COMMAND_ERROR,
 		[qr{cannot coerce boolean to int}], q{\set b random(FALSE, TRUE)}
 	],
 	[
-		'set number of args mismatch',        1,
+		'set number of args mismatch', 1, SYNTAX_ERROR,
 		[qr{unexpected number of arguments}], q{\set d ln(1.0, 2.0))}
 	],
 	[
-		'set at least one arg',               1,
+		'set at least one arg', 1, SYNTAX_ERROR,
 		[qr{at least one argument expected}], q{\set i greatest())}
 	],
 
 	# SETSHELL
 	[
-		'setshell not an int',                0,
+		'setshell not an int', 0, META_COMMAND_ERROR,
 		[qr{command must return an integer}], q{\setshell i echo -n one}
 	],
-	[ 'setshell missing arg', 1, [qr{missing argument }], q{\setshell var} ],
 	[
-		'setshell no such command',   0,
+		'setshell missing arg', 1, SYNTAX_ERROR,
+		[qr{missing argument }], q{\setshell var}
+	],
+	[
+		'setshell no such command', 0, META_COMMAND_ERROR,
 		[qr{could not read result }], q{\setshell var no-such-command}
 	],
 
 	# SLEEP
 	[
-		'sleep undefined variable',      0,
+		'sleep undefined variable', 0, META_COMMAND_ERROR,
 		[qr{sleep: undefined variable}], q{\sleep :nosuchvariable}
 	],
 	[
-		'sleep too many args',    1,
+		'sleep too many args', 1, SYNTAX_ERROR,
 		[qr{too many arguments}], q{\sleep too many args}
 	],
 	[
-		'sleep missing arg', 1,
+		'sleep missing arg', 1, SYNTAX_ERROR,
 		[ qr{missing argument}, qr{\\sleep} ], q{\sleep}
 	],
 	[
-		'sleep unknown unit',         1,
+		'sleep unknown unit', 1, SYNTAX_ERROR,
 		[qr{unrecognized time unit}], q{\sleep 1 week}
 	],
 
+	# CONDITIONAL BLOCKS
+	[   'error inside a conditional block', 0, SQL_ERROR,
+		[qr{ERROR:  division by zero}],
+		q{-- error inside a conditional block
+\if true
+SELECT 1 / 0;
+\endif
+}
+	],
+
 	# MISC
 	[
-		'misc invalid backslash command',         1,
+		'misc invalid backslash command', 1, SYNTAX_ERROR,
 		[qr{invalid command .* "nosuchcommand"}], q{\nosuchcommand}
 	],
-	[ 'misc empty script', 1, [qr{empty command list for script}], q{} ],
 	[
-		'bad boolean',                     0,
+		'misc empty script', 1, SYNTAX_ERROR,
+		[qr{empty command list for script}], q{}
+	],
+	[
+		'bad boolean', 0, META_COMMAND_ERROR,
 		[qr{malformed variable.*trueXXX}], q{\set b :badtrue or true}
 	],);
 
 
 for my $e (@errors)
 {
-	my ($name, $status, $re, $script) = @$e;
+	my ($name, $status, $error_type, $re, $script) = @$e;
 	my $n = '001_pgbench_error_' . $name;
 	$n =~ s/ /_/g;
+	my $test_name = 'pgbench script error: ' . $name;
+	my $stdout_re;
+
+	if ($status)
+	{
+		# only syntax errors get non-zero exit status
+		# internal error which should never occur
+		die $test_name . ": unexpected error type: " . $error_type . "\n"
+		if ($error_type != SYNTAX_ERROR);
+
+		$stdout_re = [ qr{^$} ];
+	}
+	else
+	{
+		$stdout_re =
+			[ qr{processed: 0/1}, qr{number of errors: 1 \(100.000%\)},
+			  qr{^((?!number of retried)(.|\n))*$} ];
+
+		if ($error_type == SQL_ERROR)
+		{
+			push @$stdout_re,
+				qr{number of serialization errors: 0 \(0.000%\)},
+				qr{number of deadlock errors: 0 \(0.000%\)},
+				qr{number of other SQL errors: 1 \(100.000%\)};
+		}
+		elsif ($error_type == META_COMMAND_ERROR)
+		{
+			push @$stdout_re,
+				qr{number of errors in meta-commands: 1 \(100.000%\)};
+		}
+		else
+		{
+			# internal error which should never occur
+			die $test_name . ": unexpected error type: " . $error_type . "\n";
+		}
+	}
+
 	pgbench(
-		'-n -t 1 -Dfoo=bla -Dnull=null -Dtrue=true -Done=1 -Dzero=0.0 -Dbadtrue=trueXXX -M prepared',
+		'-n -t 1 -Dfoo=bla -Dnull=null -Dtrue=true -Done=1 -Dzero=0.0 -Dbadtrue=trueXXX -M prepared --debug-fails --errors-detailed',
 		$status,
-		[ $status ? qr{^$} : qr{processed: 0/1} ],
+		$stdout_re,
 		$re,
-		'pgbench script error: ' . $name,
+		$test_name,
 		{ $n => $script });
 }
 
@@ -848,6 +939,209 @@ pgbench(
 check_pgbench_logs("$bdir/001_pgbench_log_3", 1, 10, 10,
 	qr{^\d \d{1,2} \d+ \d \d+ \d+$});
 
+# Test the concurrent update in the table row and deadlocks.
+
+$node->safe_psql('postgres',
+	'CREATE UNLOGGED TABLE first_client_table (value integer); '
+  . 'CREATE UNLOGGED TABLE xy (x integer, y integer); '
+  . 'INSERT INTO xy VALUES (1, 2);');
+
+# Serialization failure and retry
+
+local $ENV{PGOPTIONS} = "-c default_transaction_isolation=repeatable\\ read";
+
+# Check that we have a serialization failure and the same random value of the
+# delta variable in the next try
+my $err_pattern =
+	"(client (0|1) sending UPDATE xy SET y = y \\+ -?\\d+\\b).*"
+  . "client \\g2 got a failure in command 3 \\(SQL\\) of script 0; "
+  . "ERROR:  could not serialize access due to concurrent update\\b.*"
+  . "\\g1";
+
+pgbench(
+	"-n -c 2 -t 1 -d --max-tries 2",
+	0,
+	[ qr{processed: 2/2\b}, qr{^((?!number of errors)(.|\n))*$},
+	  qr{number of retried: 1\b}, qr{number of retries: 1\b} ],
+	[ qr/$err_pattern/s ],
+	'concurrent update with retrying',
+	{
+		'001_pgbench_serialization' => q{
+-- What's happening:
+-- The first client starts the transaction with the isolation level Repeatable
+-- Read:
+--
+-- BEGIN;
+-- UPDATE xy SET y = ... WHERE x = 1;
+--
+-- The second client starts a similar transaction with the same isolation level:
+--
+-- BEGIN;
+-- UPDATE xy SET y = ... WHERE x = 1;
+-- <waiting for the first client>
+--
+-- The first client commits its transaction, and the second client gets a
+-- serialization failure.
+
+\set delta random(-5000, 5000)
+
+-- The second client will stop here
+SELECT pg_advisory_lock(0);
+
+-- Start transaction with concurrent update
+BEGIN;
+UPDATE xy SET y = y + :delta WHERE x = 1 AND pg_advisory_lock(1) IS NOT NULL;
+
+-- Wait for the second client
+DO $$
+DECLARE
+  exists boolean;
+  waiters integer;
+BEGIN
+  -- The second client always comes in second, and the number of rows in the
+  -- table first_client_table reflect this. Here the first client inserts a row,
+  -- so the second client will see a non-empty table in the second try to run a
+  -- previously failed transaction.
+  SELECT EXISTS (SELECT * FROM first_client_table) INTO STRICT exists;
+  IF NOT exists THEN
+	-- Let the second client begin
+	PERFORM pg_advisory_unlock(0);
+	-- And wait until the second client tries to get the same lock
+	LOOP
+	  SELECT COUNT(*) INTO STRICT waiters FROM pg_locks WHERE
+	  locktype = 'advisory' AND objsubid = 1 AND
+	  ((classid::bigint << 32) | objid::bigint = 1::bigint) AND NOT granted;
+	  IF waiters = 1 THEN
+		INSERT INTO first_client_table VALUES (1);
+
+		-- Exit loop
+		EXIT;
+	  END IF;
+	END LOOP;
+  END IF;
+END$$;
+
+COMMIT;
+SELECT pg_advisory_unlock_all();
+}
+	});
+
+# Clean up
+
+$node->safe_psql('postgres', 'DELETE FROM first_client_table;');
+
+local $ENV{PGOPTIONS} = "-c default_transaction_isolation=read\\ committed";
+
+# Deadlock failure and retry
+
+# Check that we have a deadlock failure
+$err_pattern =
+	"client (0|1) got a failure in command (3|5) \\(SQL\\) of script 0; "
+  . "ERROR:  deadlock detected\\b";
+
+pgbench(
+	"-n -c 2 -t 1 --debug-fails --max-tries 2",
+	0,
+	[ qr{processed: 2/2\b}, qr{^((?!number of errors)(.|\n))*$},
+	  qr{number of retried: 1\b}, qr{number of retries: 1\b} ],
+	[ qr{$err_pattern} ],
+	'deadlock with retrying',
+	{
+		'001_pgbench_deadlock' => q{
+-- What's happening:
+-- The first client gets the lock 2.
+-- The second client gets the lock 3 and tries to get the lock 2.
+-- The first client tries to get the lock 3 and one of them gets a deadlock
+-- falure.
+--
+-- A client who has received a deadlock failure is called a (future) failed
+-- client although it will retry the failed transaction.
+--
+-- Also any successfull client must hold a lock at the transaction start so the
+-- failed client will wait until the successfull one releases all of its locks
+-- at the transaction end (we do not want any failures again).
+
+-- Since the client in the failed transaction has not released the blocking
+-- locks, let's do this here.
+SELECT pg_advisory_unlock_all();
+
+-- The second and future failed clients will stop here
+SELECT pg_advisory_lock(0);
+SELECT pg_advisory_lock(1);
+
+-- The second and future failed clients always come after the first and the
+-- number of rows in the table first_client_table reflect this. Here the first
+-- client inserts a row, so the second and future failed clients will see a
+-- non-empty table.
+DO $$
+DECLARE
+  exists boolean;
+BEGIN
+  SELECT EXISTS (SELECT * FROM first_client_table) INTO STRICT exists;
+  IF exists THEN
+	-- We are the second or the failed client
+
+	-- The first client will take care by itself of this lock (see below)
+	PERFORM pg_advisory_unlock(0);
+
+	PERFORM pg_advisory_lock(3);
+
+	-- The second client can get a deadlock here
+	PERFORM pg_advisory_lock(2);
+  ELSE
+	-- We are the first client
+
+	-- This code should not be used in a new transaction after a failure
+	INSERT INTO first_client_table VALUES (1);
+
+	PERFORM pg_advisory_lock(2);
+  END IF;
+END$$;
+
+DO $$
+DECLARE
+  num_rows integer;
+  waiters integer;
+BEGIN
+  -- Check if we are the first client
+  SELECT COUNT(*) FROM first_client_table INTO STRICT num_rows;
+  IF num_rows = 1 THEN
+	-- This code should not be used in a new transaction after a failure
+	INSERT INTO first_client_table VALUES (2);
+
+	-- Let the second client begin
+	PERFORM pg_advisory_unlock(0);
+	PERFORM pg_advisory_unlock(1);
+
+	-- Make sure the second client is ready for deadlock
+	LOOP
+	  SELECT COUNT(*) INTO STRICT waiters FROM pg_locks WHERE
+	  locktype = 'advisory' AND
+	  objsubid = 1 AND
+	  ((classid::bigint << 32) | objid::bigint = 2::bigint) AND
+	  NOT granted;
+
+	  IF waiters = 1 THEN
+	    -- Exit loop
+		EXIT;
+	  END IF;
+	END LOOP;
+
+	PERFORM pg_advisory_lock(0);
+    -- And the second client took care by itself of the lock 1
+  END IF;
+END$$;
+
+-- The first client can get a deadlock here
+SELECT pg_advisory_lock(3);
+
+SELECT pg_advisory_unlock_all();
+}
+	});
+
+# Clean up
+$node->safe_psql('postgres', 'DELETE FROM first_client_table;');
+
 # done
 $node->stop;
 done_testing();
diff --git a/src/bin/pgbench/t/002_pgbench_no_server.pl b/src/bin/pgbench/t/002_pgbench_no_server.pl
index c1c2c1e..50626c2 100644
--- a/src/bin/pgbench/t/002_pgbench_no_server.pl
+++ b/src/bin/pgbench/t/002_pgbench_no_server.pl
@@ -157,6 +157,11 @@ my @options = (
 			qr{error while setting random seed from --random-seed option}
 		]
 	],
+	[
+		'bad maximum number of tries',
+		'--max-tries -10',
+		[ qr{invalid number of maximum tries: "-10"} ]
+	],
 
 	# loging sub-options
 	[
diff --git a/src/fe_utils/conditional.c b/src/fe_utils/conditional.c
index db2a0a5..4d14066 100644
--- a/src/fe_utils/conditional.c
+++ b/src/fe_utils/conditional.c
@@ -24,13 +24,25 @@ conditional_stack_create(void)
 }
 
 /*
- * destroy stack
+ * Destroy all the elements from the stack. The stack itself is not freed.
  */
 void
-conditional_stack_destroy(ConditionalStack cstack)
+conditional_stack_reset(ConditionalStack cstack)
 {
+	if (!cstack)
+		return;					/* nothing to do here */
+
 	while (conditional_stack_pop(cstack))
 		continue;
+}
+
+/*
+ * destroy stack
+ */
+void
+conditional_stack_destroy(ConditionalStack cstack)
+{
+	conditional_stack_reset(cstack);
 	free(cstack);
 }
 
diff --git a/src/include/fe_utils/conditional.h b/src/include/fe_utils/conditional.h
index 9b91de5..59c8d8a 100644
--- a/src/include/fe_utils/conditional.h
+++ b/src/include/fe_utils/conditional.h
@@ -73,6 +73,8 @@ typedef struct ConditionalStackData *ConditionalStack;
 
 extern ConditionalStack conditional_stack_create(void);
 
+extern void conditional_stack_reset(ConditionalStack cstack);
+
 extern void conditional_stack_destroy(ConditionalStack cstack);
 
 extern int	conditional_stack_depth(ConditionalStack cstack);
-- 
2.7.4

